from collections import defaultdict


def group_by(table, column):
    """
    Groups rows in the table by the value in the specified column.

    Parameters:
        table (list of dict): The table to group, as a list of dictionaries.
        column (str): The column name to group by.

    Returns:
        list of list of dict: A list of groups, each group is a list of rows (dicts) sharing the same value in the specified column.
    """
    groups = defaultdict(list)
    for row in table:
        groups[row[column]].append(row)
    return list(groups.values())


def pivot(table, index, columns, values):
    """
    Pivots the table to reorganize data, similar to a spreadsheet pivot.

    Parameters:
        table (list of dict): The table to pivot, as a list of dictionaries.
        index (str): The column to use as the new row index.
        columns (str): The column whose unique values will become new columns.
        values (str): The column whose values fill the new table.

    Returns:
        list of dict: The pivoted table as a list of dictionaries.
    """
    result = defaultdict(dict)
    for row in table:
        row_idx = row[index]
        col_key = row[columns]
        val = row[values]
        result[row_idx][col_key] = val
    # Convert to list of dicts, including index as a column
    return [dict({index: idx}, **cols) for idx, cols in result.items()]


def melt(table, id_vars, value_vars, var_name, value_name):
    """
    Unpivots the table from wide to long format.

    Parameters:
        table (list of dict): The table to melt, as a list of dictionaries.
        id_vars (list of str): Columns to use as identifier variables.
        value_vars (list of str): Columns to unpivot.
        var_name (str): Name to use for the variable column.
        value_name (str): Name to use for the value column.

    Returns:
        list of dict: The melted table as a list of dictionaries in long format.
    """
    result = []
    for row in table:
        for col in value_vars:
            melted = {k: row[k] for k in id_vars}
            melted[var_name] = col
            melted[value_name] = row[col]
            result.append(melted)
    return result



from __future__ import annotations

import pathlib
from typing import Callable

from lark import Lark
from lark import Transformer

from aeon.core.substitutions import liquefy
from aeon.core.terms import Abstraction
from aeon.core.terms import Annotation
from aeon.core.terms import Application
from aeon.core.terms import Hole
from aeon.core.terms import If
from aeon.core.terms import Let
from aeon.core.terms import Literal
from aeon.core.terms import Rec
from aeon.core.terms import Term
from aeon.core.terms import TypeAbstraction
from aeon.core.terms import TypeApplication
from aeon.core.terms import Var
from aeon.core.types import AbstractionType, TypeConstructor
from aeon.core.types import BaseKind
from aeon.core.types import RefinedType
from aeon.core.types import StarKind
from aeon.core.types import t_bool
from aeon.core.types import t_float
from aeon.core.types import t_int
from aeon.core.types import t_string
from aeon.core.types import top
from aeon.core.types import Type
from aeon.core.types import TypePolymorphism
from aeon.core.types import TypeVar
from aeon.utils.ast_helpers import i0
from aeon.utils.ast_helpers import mk_binop
from aeon.utils.name import Name


class TreeToCore(Transformer):
    counter: int

    def __init__(self, start_counter=0):
        self.counter = start_counter

    def fresh(self) -> Name:
        self.counter += 1
        return Name("anf", self.counter)

    def same(self, args):
        return args[0]

    # Types
    def refined_t(self, args):
        return RefinedType(Name(args[0]), args[1], liquefy(args[2]))

    def abstraction_t(self, args):
        return AbstractionType(Name(args[0]), args[1], args[2])

    def polymorphism_t(self, args):
        return TypePolymorphism(Name(args[0]), args[1], args[2])

    def simple_t(self, args):
        n = str(args[0])
        if n == "Top":
            return top
        elif n in ["Unit", "Int", "Bool", "Float", "String"]:
            return TypeConstructor(Name(n, 0))
        else:
            return TypeVar(Name(n))

    def constructor_t(self, args):
        return TypeConstructor(Name(args[0]), args[1:])

    # Expressions

    def annotation(self, args):
        return Annotation(args[0], args[1])

    def minus(self, args):
        if isinstance(args[0], Literal) and args[0].type == t_int:
            return Literal(-args[0].value, args[0].type)
        return mk_binop(lambda: self.fresh(), "-", i0, args[0])

    def let_e(self, args):
        return Let(Name(args[0]), args[1], args[2])

    def rec_e(self, args):
        return Rec(Name(args[0]), args[1], args[2], args[3])

    def if_e(self, args):
        return If(args[0], args[1], args[2])

    def nnot(self, args):
        return Application(Var(Name("!", 0)), args[0])

    def binop_eq(self, args):
        return self.binop(args, "==")

    def binop_neq(self, args):
        return self.binop(args, "!=")

    def binop_and(self, args):
        return self.binop(args, "&&")

    def binop_or(self, args):
        return self.binop(args, "||")

    def binop_lt(self, args):
        return self.binop(args, "<")

    def binop_le(self, args):
        return self.binop(args, "<=")

    def binop_gt(self, args):
        return self.binop(args, ">")

    def binop_ge(self, args):
        return self.binop(args, ">=")

    def binop_imp(self, args):
        return self.binop(args, "-->")

    def binop_plus(self, args):
        return self.binop(args, "+")

    def binop_minus(self, args):
        return self.binop(args, "-")

    def binop_mult(self, args):
        return self.binop(args, "*")

    def binop_div(self, args):
        return self.binop(args, "/")

    def binop_mod(self, args):
        return self.binop(args, "%")

    def binop(self, args, op):
        return mk_binop(lambda: self.fresh(), op, args[0], args[1])

    def application_e(self, args):
        return Application(args[0], args[1])

    def abstraction_e(self, args):
        return Abstraction(Name(args[0]), args[1])

    def tabstraction_e(self, args):
        return TypeAbstraction(Name(args[0]), args[1], args[2])

    def type_application_e(self, args):
        return TypeApplication(args[0], args[1])

    def var(self, args):
        return Var(Name(str(args[0]).strip()))

    def hole(self, args):
        return Hole(Name(args[0]))

    def int_lit(self, args):
        return Literal(int(args[0]), type=t_int)

    def float_lit(self, args):
        return Literal(float(args[0]), type=t_float)

    def bool_lit(self, args):
        value = str(args[0]) == "true"
        return Literal(value, type=t_bool)

    def string_lit(self, args):
        return Literal(args[0], type=t_string)

    def ESCAPED_STRING(self, val):
        # TODO: This is terrible and doesn't handle escapes
        v = str(val)[1:-1]
        return v

    def base_kind(self, args):
        return BaseKind()

    def star_kind(self, args):
        return StarKind()


def mk_parser(rule="start", start_counter=0):
    return Lark.open(
        pathlib.Path(__file__).parent.absolute() / "aeon_core.lark",
        parser="lalr",
        # lexer='standard',
        start=rule,
        transformer=TreeToCore(start_counter),
    )


parse_type: Callable[[str], Type] = mk_parser("type").parse
parse_term: Callable[[str], Term] = mk_parser("expression").parse



from __future__ import annotations

from aeon.core.liquid import LiquidVar
from aeon.core.liquid_ops import mk_liquid_and
from aeon.core.substitutions import substitution_in_liquid
from aeon.core.types import AbstractionType, TypeConstructor
from aeon.core.types import RefinedType
from aeon.core.types import Type
from aeon.core.types import TypePolymorphism
from aeon.core.types import TypeVar
from aeon.utils.name import Name


def type_substitution(t: Type, alpha: Name, beta: Type) -> Type:
    """t[alpha := beta], standard substition."""
    assert isinstance(t, Type)

    def rec(x):
        return type_substitution(x, alpha, beta)

    match t:
        case TypeVar(name):
            if name == alpha:
                return beta
            else:
                return t
        case RefinedType(name, ity, ref, loc):
            match rec(ity):
                case RefinedType(iname, iity, iref) as city:
                    return RefinedType(
                        name, iity, mk_liquid_and(ref, substitution_in_liquid(iref, LiquidVar(name), iname)), loc=loc
                    )
                case AbstractionType(_, _, _):
                    assert False, f"Abstraction types cannot be refined: {t} to {ity} to {rec(ity)}"
                case city:
                    return RefinedType(name, city, ref, loc=loc)
        case AbstractionType(aname, aty, rty, loc):
            return AbstractionType(aname, rec(aty), rec(rty), loc=loc)
        case TypePolymorphism(name, kind, body, loc):
            if name == alpha:
                return t
            else:
                return TypePolymorphism(name, kind, rec(body), loc=loc)
        case TypeConstructor(name, args, loc):
            return TypeConstructor(name, [rec(arg) for arg in args], loc=loc)
        case _:
            assert False, f"Not considered: {t} ({type(t)})"


def type_variable_instantiation(t: Type, alpha: Name, beta: Type) -> Type:
    """t[alpha |-> beta], instantiation."""

    def rec(x):
        return type_variable_instantiation(x, alpha, beta)

    if isinstance(t, TypeConstructor):
        return t
    elif isinstance(t, TypeVar) and t.name == alpha:
        return beta
    elif isinstance(t, TypeVar) and t.name != alpha:
        return t
    elif (
        isinstance(t, RefinedType)
        and isinstance(t.type, TypeVar)
        and t.type.name == alpha
        and isinstance(beta, RefinedType)
    ):
        return RefinedType(
            t.name,
            beta.type,
            mk_liquid_and(
                t.refinement,
                substitution_in_liquid(beta.refinement, LiquidVar(t.name), beta.name),
            ),
            t.loc,
        )
    elif isinstance(t, RefinedType):
        return RefinedType(t.name, rec(t.type), t.refinement, t.loc)
    elif isinstance(t, AbstractionType):
        return AbstractionType(t.var_name, rec(t.var_type), rec(t.type), t.loc)
    elif isinstance(t, TypePolymorphism):
        return TypePolymorphism(t.name, t.kind, rec(t.body), t.loc)
    else:
        assert False



from __future__ import annotations

from dataclasses import dataclass
from typing import Generator

from aeon.core.liquid import LiquidApp
from aeon.core.liquid import LiquidLiteralBool
from aeon.core.liquid import LiquidLiteralInt
from aeon.core.liquid import LiquidLiteralString
from aeon.core.liquid import LiquidTerm
from aeon.core.liquid import LiquidVar
from aeon.core.types import AbstractionType, Top, TypeVar
from aeon.core.types import TypeConstructor
from aeon.utils.name import Name


class Constraint:
    pass


@dataclass
class LiquidConstraint(Constraint):
    expr: LiquidTerm

    def __repr__(self):
        return repr(self.expr)


@dataclass
class Conjunction(Constraint):
    c1: Constraint
    c2: Constraint

    def __repr__(self):
        return f"({self.c1}) ∧ ({self.c2})"


@dataclass
class UninterpretedFunctionDeclaration(Constraint):
    name: Name
    type: AbstractionType
    seq: Constraint

    def __repr__(self):
        return f"fun {self.name}:{self.type} => {self.seq}"


@dataclass
class Implication(Constraint):
    name: Name
    base: TypeConstructor | TypeVar | Top
    pred: LiquidTerm
    seq: Constraint

    def __post_init__(self):
        assert isinstance(self.name, Name)

    def __repr__(self):
        return f"∀{self.name}:{self.base}, ({self.pred}) => {self.seq}"


@dataclass
class TypeVarDeclaration(Constraint):
    name: Name
    seq: Constraint

    def __repr__(self):
        return f"type {self.name}, {self.seq}"


def variables_in_liq(t: LiquidTerm) -> Generator[str, None, None]:
    if isinstance(t, LiquidLiteralBool) or isinstance(t, LiquidLiteralInt) or isinstance(t, LiquidLiteralString):
        pass
    elif isinstance(t, LiquidVar):
        yield str(t.name)
    elif isinstance(t, LiquidApp):
        yield str(t.fun)
        for a in t.args:
            yield from variables_in_liq(a)


def variables_free_in(c: Constraint) -> Generator[str, None, None]:
    if isinstance(c, Conjunction):
        yield from variables_free_in(c.c1)
        yield from variables_free_in(c.c2)
    elif isinstance(c, Implication):
        for k in variables_in_liq(c.pred):
            if k != c.name:
                yield k
        for k in variables_free_in(c.seq):
            if k != c.name:
                yield k
    elif isinstance(c, LiquidConstraint):
        yield from variables_in_liq(c.expr)



from __future__ import annotations

from aeon.utils.name import Name
from aeon.core.liquid import LiquidVar
from aeon.core.substitutions import substitution_in_liquid
from aeon.core.types import AbstractionType
from aeon.core.types import TypeConstructor
from aeon.core.types import extract_parts
from aeon.core.types import TypePolymorphism
from aeon.core.types import TypeVar
from aeon.typechecking.context import TypeConstructorBinder
from aeon.typechecking.context import TypeBinder
from aeon.typechecking.context import TypingContext
from aeon.typechecking.context import UninterpretedBinder
from aeon.typechecking.context import VariableBinder
from aeon.verification.horn import solve
from aeon.verification.sub import is_first_order_function
from aeon.verification.vcs import Constraint
from aeon.verification.vcs import Implication
from aeon.verification.vcs import UninterpretedFunctionDeclaration

from aeon.core.liquid_ops import ops


def entailment_context(ctx: TypingContext, c: Constraint) -> Constraint:
    for entry in ctx.entries[::-1]:
        match entry:
            case VariableBinder(name, AbstractionType(vname, vtype, rtype)):
                if name in ops:
                    pass
                elif is_first_order_function(AbstractionType(vname, vtype, rtype)):
                    c = UninterpretedFunctionDeclaration(name, AbstractionType(vname, vtype, rtype), c)
                else:
                    pass
            case VariableBinder(name, TypePolymorphism(_, _, _)):
                if name in ops:
                    pass
                else:
                    # TODO: polymorphism
                    # Right now we are ignoring lifting functions with polymorphism
                    pass
            case VariableBinder(name, ty):
                (nname, base, cond) = extract_parts(ty)
                match base:
                    case TypeConstructor(_, []) | TypeVar(_):
                        ncond = substitution_in_liquid(cond, LiquidVar(name), nname)
                        c = Implication(name, base, ncond, c)
                    case TypeConstructor(_, _):
                        ncond = substitution_in_liquid(cond, LiquidVar(name), nname)
                        c = Implication(name, TypeConstructor(Name("Int", 0), []), ncond, c)
                    case _:
                        assert False, f"Unknown base: {base}"
            case TypeBinder(name, _):
                pass
                # TODO: Consider passing as a concrete placeholder type for SMT
            case UninterpretedBinder(name, type):
                c = UninterpretedFunctionDeclaration(name, type, c)
            case TypeConstructorBinder(name, _):
                pass
            case _:
                assert False, f"Untreated {ctx}."
    return c


def entailment(ctx: TypingContext, c: Constraint) -> bool:
    c = entailment_context(ctx, c)
    return solve(c)



from __future__ import annotations

from loguru import logger

from aeon.core.liquid import LiquidLiteralBool
from aeon.core.liquid import LiquidVar
from aeon.core.substitutions import substitution_in_liquid
from aeon.core.substitutions import substitution_in_type
from aeon.core.terms import Var
from aeon.core.types import AbstractionType, TypeConstructor, TypeVar
from aeon.core.types import RefinedType
from aeon.core.types import Top
from aeon.core.types import Type
from aeon.core.types import TypePolymorphism
from aeon.core.types import t_unit
from aeon.typechecking.context import TypingContext
from aeon.verification.vcs import Conjunction, UninterpretedFunctionDeclaration
from aeon.verification.vcs import Constraint
from aeon.verification.vcs import Implication
from aeon.verification.vcs import LiquidConstraint
from aeon.utils.name import Name, fresh_counter

ctrue = LiquidConstraint(LiquidLiteralBool(True))
cfalse = LiquidConstraint(LiquidLiteralBool(False))


def ensure_refined(t: Type) -> Type:
    """Ensures that the Base Types and TypeVars are refined. All other types remain the same."""
    match t:
        case RefinedType(_, _, _):
            return t
        case TypeVar(Name(name)):
            return RefinedType(Name(f"singleton_{name}", fresh_counter.fresh()), t, LiquidLiteralBool(True))
        case TypeConstructor(Name(name), _):
            return RefinedType(Name(f"singleton_{name}", fresh_counter.fresh()), t, LiquidLiteralBool(True))
        case _:
            return t


def is_first_order_function(at: AbstractionType):
    v: Type = at
    while isinstance(v, AbstractionType):
        match v.var_type:
            case AbstractionType(_, _, _):
                return False
            case Top() | RefinedType(_, _, _) | TypeVar(_) | TypeConstructor(_, _):
                pass
            case _:
                assert False
        v = v.type
    return True


def lower_constraint_type(ttype: Type) -> Type:
    match ttype:
        case TypeVar(name):
            return TypeConstructor(name)
        case Top():
            return t_unit
        case AbstractionType(name, b, r):
            return AbstractionType(name, lower_constraint_type(b), lower_constraint_type(r))
        case RefinedType(_, t, _):
            return lower_constraint_type(t)
        case TypeConstructor(name, args):
            if args:
                # Polymorphic types are represented by top
                return TypeConstructor(Name("Top", 0))
            else:
                return TypeConstructor(name)

        case _:
            assert False, f"Unsupport type in constraint {ttype} ({type(ttype)})"


def implication_constraint(name: Name, ty: Type, c: Constraint) -> Constraint:
    match ty:
        case Top() | TypeVar(_) | TypeConstructor(_, _):
            basety = lower_constraint_type(ty)
            assert isinstance(basety, TypeConstructor)
            return Implication(name, basety, LiquidLiteralBool(True), c)
        case RefinedType(tname, ttype, tref):
            ref_subs = substitution_in_liquid(tref, LiquidVar(name), tname)
            ltype = lower_constraint_type(ttype)
            assert isinstance(ltype, TypeConstructor) or isinstance(ltype, TypeVar)
            return Implication(name, ltype, ref_subs, c)
        case AbstractionType(_, _, _):
            # TODO Poly Refl: instead of true, reflect the implementation of the function?
            if is_first_order_function(ty):
                absty = lower_constraint_type(ty)
                assert isinstance(absty, AbstractionType)
                return UninterpretedFunctionDeclaration(name, absty, c)
            else:
                return c
        case TypePolymorphism(_, _, _):
            return c
        case _:
            assert False


def sub(ctx: TypingContext, t1: Type, t2: Type) -> Constraint:
    if t2 == Top():
        return ctrue
    match (ensure_refined(t1), ensure_refined(t2)):
        case RefinedType(n1, ty1, r1), RefinedType(n2, ty2, r2):
            if ty2 == Top():
                return ctrue
            if ty1 != ty2:
                return cfalse

            new_name: Name = Name(n1.name + n2.name, fresh_counter.fresh())

            # Performs substition on t2 to have the same name of t1
            r2_ = substitution_in_liquid(r2, LiquidVar(new_name), n2)
            r1_ = substitution_in_liquid(r1, LiquidVar(new_name), n1)
            lowert = lower_constraint_type(ty1)
            assert isinstance(lowert, TypeConstructor)
            rconstraint = Implication(new_name, lowert, r1_, LiquidConstraint(r2_))

            return rconstraint
        case TypePolymorphism(_, _, _), _:
            return ctrue
        case AbstractionType(a1, t1, rt1), AbstractionType(a2, t2, rt2):
            new_name_a: Name = Name(a1.name + a2.name, fresh_counter.fresh())

            c0 = sub(ctx, t2, t1)
            rt1_ = substitution_in_type(rt1, Var(new_name_a), a1)
            rt2_ = substitution_in_type(rt2, Var(new_name_a), a2)
            c1 = sub(ctx, rt1_, rt2_)
            return Conjunction(c0, implication_constraint(new_name_a, t2, c1))
        case TypeConstructor(name1, args1), TypeConstructor(name2, args2):
            if name1 != name2:
                return cfalse
            if len(args1) != len(args2):
                return cfalse
            for it1, it2 in zip(args1, args2):
                if it1 != it2:  # TODO polytypes: subtyping here?
                    return cfalse
            return ctrue
        case _:
            logger.error(f"Failed subtyping by exhaustion: {t1} <: {t2}")
            return cfalse



from __future__ import annotations

from aeon.core.liquid import LiquidApp
from aeon.core.types import LiquidHornApplication, TypeConstructor, TypePolymorphism
from aeon.core.liquid import LiquidLiteralBool
from aeon.core.liquid import LiquidLiteralFloat
from aeon.core.liquid import LiquidLiteralInt
from aeon.core.liquid import LiquidLiteralString
from aeon.core.liquid import LiquidTerm
from aeon.core.liquid import LiquidVar
from aeon.core.terms import Abstraction, TypeAbstraction, TypeApplication
from aeon.core.terms import Annotation
from aeon.core.terms import Application
from aeon.core.terms import Hole
from aeon.core.terms import If
from aeon.core.terms import Let
from aeon.core.terms import Literal
from aeon.core.terms import Rec
from aeon.core.terms import Term
from aeon.core.terms import Var
from aeon.core.types import AbstractionType
from aeon.core.types import RefinedType
from aeon.core.types import Top
from aeon.core.types import Type
from aeon.core.types import TypeVar
from aeon.utils.name import Name


def substitute_vartype(t: Type, rep: Type, name: Name) -> Type:
    """Replaces all occurrences of vartypes name in t by rep."""

    def rec(k: Type):
        return substitute_vartype(k, rep, name)

    match t:
        case TypeVar(tname):
            if tname == name:
                return rep
            else:
                return t
        case RefinedType(rname, ty, ref, loc):
            assert not isinstance(ty, RefinedType)
            return RefinedType(rname, rec(ty), ref, loc=loc)
        case AbstractionType(a, aty, rty, loc):
            return AbstractionType(a, rec(aty), rec(rty), loc=loc)
        case TypePolymorphism(pname, kind, body, loc):
            if name == pname:
                return t
            else:
                return TypePolymorphism(pname, kind, rec(body), loc=loc)
        case TypeConstructor(cname, args, loc):
            return TypeConstructor(cname, [rec(arg) for arg in args], loc=loc)
        case _:
            assert False, f"type {t} ({type(t)}) not allows in substitution."


def substitute_vartype_in_term(t: Term, rep: Type, name: Name) -> Term:
    def rec(x: Term):
        return substitute_vartype_in_term(x, rep, name)

    match t:
        case Literal():
            return t
        case Var():
            return t
        case Hole():
            return t
        case Application(fun, arg, loc):
            return Application(fun=rec(fun), arg=rec(arg), loc=loc)
        case Abstraction(var_name, body, loc):
            return Abstraction(var_name, rec(body), loc=loc)
        case Let(var_name, var_value, body, loc):
            n_value = rec(var_value)
            n_body = rec(body)
            return Let(var_name, n_value, n_body, loc=loc)
        case Rec(var_name, var_type, var_value, body, loc):
            n_value = rec(var_value)
            n_type = substitute_vartype(var_type, rep, name)
            n_body = rec(body)
            return Rec(var_name, n_type, n_value, n_body, loc=loc)
        case Annotation(expr, type, loc):
            n_type = substitute_vartype(type, rep, name)
            return Annotation(rec(expr), n_type, loc=loc)
        case If(cond, then, otherwise, loc):
            n_cond = rec(cond)
            n_then = rec(then)
            n_otherwise = rec(otherwise)
            return If(n_cond, n_then, n_otherwise, loc=loc)
        case _:
            assert False


def substitution_in_liquid(
    t: LiquidTerm,
    rep: LiquidTerm,
    name: Name,
) -> LiquidTerm:
    """substitutes name in the term t with the new replacement term rep."""
    match t:
        case LiquidLiteralBool(_) | LiquidLiteralInt(_) | LiquidLiteralFloat(_) | LiquidLiteralString(_):
            return t
        case LiquidVar(tname):
            if tname == name:
                return rep
            else:
                return t
        case LiquidApp(aname, args, loc):
            if aname == name:
                assert isinstance(rep, LiquidVar)
                nname = rep.name
            else:
                nname = aname
            return LiquidApp(nname, [substitution_in_liquid(a, rep, name) for a in args], loc=loc)

        case LiquidHornApplication(aname, argtypes, loc):
            if aname == name:
                assert isinstance(rep, LiquidVar)
                nname = rep.name
            else:
                nname = aname
            return LiquidHornApplication(
                aname, [(substitution_in_liquid(a, rep, name), t) for (a, t) in argtypes], loc=loc
            )
        case _:
            assert False, f"{t} ({type(t)}) not allowed in substitution."


def substitution_liquid_in_type(t: Type, rep: LiquidTerm, name: Name) -> Type:
    def rec(t: Type) -> Type:
        return substitution_liquid_in_type(t, rep, name)

    match t:
        case Top() | TypeConstructor(_) | TypeVar(_):
            return t
        case AbstractionType(aname, atype, rtype, loc):
            if aname == name:
                return t
            else:
                return AbstractionType(aname, rec(atype), rec(rtype), loc=loc)
        case RefinedType(vname, ity, ref, loc):
            if name == vname:
                return t
            else:
                return RefinedType(vname, ity, substitution_in_liquid(ref, rep, name), loc=loc)
        case TypePolymorphism(name, kind, body, loc):
            return TypePolymorphism(name, kind, rec(body), loc=loc)
        case TypeConstructor(name, args, loc):
            return TypeConstructor(name, [rec(arg) for arg in args], loc=loc)
        case _:
            assert False, f"{t} not allowed"


def substitution_in_type(t: Type, rep: Term, name: Name) -> Type:
    """Substitutes name in type t with the new replacement term rep."""
    replacement: LiquidTerm | None = liquefy(rep)
    if replacement is None:
        return t

    def rec(t: Type) -> Type:
        return substitution_in_type(t, rep, name)

    match t:
        case Top() | TypeConstructor(_) | TypeVar(_):
            return t
        case AbstractionType(aname, atype, rtype, loc):
            if aname == name:
                return t
            else:
                return AbstractionType(aname, rec(atype), rec(rtype), loc=loc)
        case RefinedType(vname, ity, ref, loc):
            if name == vname:
                return t
            else:
                return RefinedType(vname, ity, substitution_in_liquid(ref, replacement, name), loc=loc)
        case TypePolymorphism(name, kind, body, loc):
            return TypePolymorphism(name, kind, rec(body), loc=loc)
        case TypeConstructor(name, args, loc):
            return TypeConstructor(name, [rec(arg) for arg in args], loc=loc)
        case _:
            assert False, f"{t} not allowed"


def substitution(t: Term, rep: Term, name: Name) -> Term:
    """Substitutes name in term t with the new replacement term rep."""

    def rec(x: Term):
        return substitution(x, rep, name)

    match t:
        case Literal(_):
            return t
        case Var(tname) | Hole(tname):
            if tname == name:
                return rep
            else:
                return t
        case Application(fun, arg, loc):
            return Application(fun=rec(fun), arg=rec(arg), loc=loc)
        case Abstraction(vname, body, loc):
            if vname == name:
                return t
            else:
                return Abstraction(vname, rec(t.body), loc=loc)
        case Let(tname, val, body, loc):
            if tname == name:
                n_value = val
                n_body = body
            else:
                n_value = rec(val)
                n_body = rec(body)
            return Let(tname, n_value, n_body, loc=loc)
        case Rec(tname, ty, val, body, loc):
            if tname == name:
                n_value = val
                n_body = body
            else:
                n_value = rec(val)
                n_body = rec(body)
            return Rec(tname, ty, n_value, n_body, loc=loc)
        case Annotation(body, ty, loc):
            return Annotation(rec(body), ty, loc=loc)
        case If(cond, then, otherwise, loc):
            return If(rec(cond), rec(then), rec(otherwise), loc=loc)
        case TypeApplication(expr, ty, loc):
            return TypeApplication(rec(expr), ty, loc=loc)
        case TypeAbstraction(pname, kind, body, loc):
            return TypeAbstraction(pname, kind, rec(body), loc=loc)
        case _:
            assert False, f"{t} not supported."


def inline_lets(t: Term) -> Term:
    """Inlines all lets in the term t."""
    match t:
        case Let(var_name, var_value, body):
            return inline_lets(substitution(body, var_value, var_name))
        case Rec(var_name, _, var_value, body):
            return inline_lets(substitution(body, var_value, var_name))
        case Application(Annotation(Abstraction(var_name, body), _), arg):
            return inline_lets(substitution(body, arg, var_name))
        case Application(Abstraction(var_name, body), arg):
            return inline_lets(substitution(body, arg, var_name))
        case Application(fun, arg, loc):
            return Application(inline_lets(fun), inline_lets(arg), loc=loc)
        case Abstraction(var_name, body, loc):
            return Abstraction(var_name, inline_lets(body), loc=loc)
        case If(cond, then, otherwise, loc):
            return If(inline_lets(cond), inline_lets(then), inline_lets(otherwise), loc=loc)
        case TypeApplication(TypeAbstraction(name, _, body), ty, loc):
            return inline_lets(substitute_vartype_in_term(body, ty, name))
        case TypeApplication(expr, ty, loc):
            return TypeApplication(inline_lets(expr), ty, loc=loc)
        case _:
            return t


def uncurry(t: Term, args: list[LiquidTerm]) -> LiquidTerm | None:
    """Uncurries the term t."""
    match t:
        case TypeApplication(body, _):
            return uncurry(body, args)
        case Application(fun, arg):
            liquid_arg = liquefy(arg)
            if liquid_arg is None:
                return None
            else:
                return uncurry(fun, [liquid_arg] + args)
        case _:
            liquid_t = liquefy(t)
            match liquid_t:
                case None:
                    return None
                case LiquidVar(name, loc):
                    return LiquidApp(name, args, loc=loc)
                case _:
                    assert False, f"Unknown term {t} {type(t)} in uncurry: {liquid_t}"


def liquefy_app(app: Application) -> LiquidApp | None:
    assert isinstance(app, Application), "not application"
    lapp = uncurry(app, [])
    assert isinstance(lapp, LiquidApp) or lapp is None
    return lapp


def liquefy_rec(t: Rec) -> LiquidTerm | None:
    value = liquefy(t.var_value)  # TODO: induction?
    body = liquefy(t.body)
    if value and body:
        return substitution_in_liquid(body, value, t.var_name)
    return None


def liquefy_let(t: Let) -> LiquidTerm | None:
    value = liquefy(t.var_value)
    body = liquefy(t.body)
    if value and body:
        return substitution_in_liquid(body, value, t.var_name)
    return None


def liquefy_if(t: If) -> LiquidTerm | None:
    co = liquefy(t.cond)
    th = liquefy(t.then)
    ot = liquefy(t.otherwise)
    if co and th and ot:
        return LiquidApp(Name("ite", 0), [co, th, ot], loc=t.loc)
    return None


def liquefy_ann(t: Annotation) -> LiquidTerm | None:
    return liquefy(t.expr)


# patterm matching term
def liquefy(rep: Term) -> LiquidTerm | None:
    """Converts a term to a liquid term."""

    rep = inline_lets(rep)

    assert isinstance(rep, Term), "not term"
    match rep:
        case Literal(val, TypeConstructor(Name("Int", 0)), loc):
            assert isinstance(val, int)
            return LiquidLiteralInt(val, loc=loc)
        case Literal(val, TypeConstructor(Name("Float", 0)), loc):
            assert isinstance(val, float)
            return LiquidLiteralFloat(val, loc=loc)
        case Literal(val, TypeConstructor(Name("Bool", 0)), loc):
            assert isinstance(val, bool)
            return LiquidLiteralBool(val, loc=loc)
        case Literal(val, TypeConstructor(Name("String", 0)), loc):
            assert isinstance(val, str)
            return LiquidLiteralString(val, loc=loc)
        case Application(_, _):
            return liquefy_app(rep)
        case TypeAbstraction(_, _, body):
            return liquefy(body)
        case Var(name, loc):
            return LiquidVar(name, loc=loc)
        case Hole(name, loc):
            return LiquidHornApplication(name, argtypes=[], loc=loc)
        case Let(_, _, _):
            return liquefy_let(rep)
        case Rec(_, _, _, _):
            return liquefy_rec(rep)
        case If(_, _, _):
            return liquefy_if(rep)
        case Annotation(_, _):
            return liquefy_ann(rep)
        case Abstraction(_, _):
            return None
        case _:
            raise Exception(f"Unable to liquefy {rep} {type(rep)}")



from __future__ import annotations

from aeon.core.liquid import LiquidApp
from aeon.core.liquid import LiquidLiteralBool
from aeon.core.liquid import LiquidTerm
from aeon.core.types import TypeConstructor, TypeVar, t_bool, t_int
from aeon.utils.name import Name


def tv(name: str) -> TypeVar:
    return TypeVar(Name(name))


liquid_prelude: dict[Name, list[TypeConstructor | TypeVar]] = {
    Name("==", 0): [tv("a"), tv("a"), t_bool],
    Name("!=", 0): [tv("a"), tv("a"), t_bool],
    Name("<", 0): [tv("a"), tv("a"), t_bool],  # TODO typeclasses: order
    Name("<=", 0): [tv("a"), tv("a"), t_bool],
    Name(">", 0): [tv("a"), tv("a"), t_bool],
    Name(">=", 0): [tv("a"), tv("a"), t_bool],
    Name("-->", 0): [t_bool, t_bool, t_bool],
    Name("&&", 0): [t_bool, t_bool, t_bool],
    Name("||", 0): [t_bool, t_bool, t_bool],
    Name("+", 0): [tv("a"), tv("a"), tv("a")],
    Name("-", 0): [tv("a"), tv("a"), tv("a")],
    Name("*", 0): [tv("a"), tv("a"), tv("a")],
    Name("/", 0): [tv("a"), tv("a"), tv("a")],
    Name("%", 0): [t_int, t_int, t_int],
    Name("!", 0): [t_bool, t_bool],
}

ops = [x for x in liquid_prelude]


def mk_liquid_and(e1: LiquidTerm, e2: LiquidTerm):
    if e1 == LiquidLiteralBool(True):
        return e2
    elif e2 == LiquidLiteralBool(True):
        return e1
    elif e1 == LiquidLiteralBool(False):
        return e1
    elif e2 == LiquidLiteralBool(False):
        return e2
    else:
        return LiquidApp(Name("&&", 0), [e1, e2])



from __future__ import annotations

from functools import reduce
from typing import Generator

from aeon.core.liquid import liquid_free_vars
from aeon.core.liquid import LiquidApp
from aeon.core.liquid import LiquidLiteralBool
from aeon.core.liquid import LiquidTerm
from aeon.core.liquid import LiquidVar
from aeon.core.substitutions import liquefy
from aeon.core.substitutions import substitution_in_liquid
from aeon.core.types import AbstractionType, TypeConstructor, Top, TypeVar
from aeon.core.types import t_bool
from aeon.frontend.parser import parse_term
from aeon.verification.smt import base_functions
from aeon.verification.vcs import Conjunction
from aeon.verification.vcs import Constraint
from aeon.verification.vcs import Implication
from aeon.verification.vcs import LiquidConstraint
from aeon.verification.vcs import UninterpretedFunctionDeclaration
from aeon.utils.name import Name, fresh_counter


def parse_liquid(t: str) -> LiquidTerm | None:
    tp = parse_term(t)
    tl = liquefy(tp)
    return tl


def imp(a: str | LiquidTerm, b: Constraint) -> Constraint:
    e = a if isinstance(a, LiquidTerm) else parse_liquid(a)
    assert e is not None
    return Implication(Name("_", fresh_counter.fresh()), t_bool, e, b)


def conj(a: Constraint, b: Constraint) -> Constraint:
    return Conjunction(a, b)


def end(a: str | LiquidTerm) -> LiquidConstraint:
    e = a if isinstance(a, LiquidTerm) else parse_liquid(a)
    assert e is not None
    return LiquidConstraint(e)


def constraint_builder(vs: list[tuple[Name, TypeConstructor | TypeVar | AbstractionType | Top]], exp: Constraint):
    for n, t in vs[::-1]:
        if isinstance(t, AbstractionType):
            exp = UninterpretedFunctionDeclaration(n, t, exp)
        else:
            exp = Implication(n, t, LiquidLiteralBool(True), exp)
    return exp


def simplify_is_true(c: Constraint):
    return isinstance(c, LiquidConstraint) and c.expr == LiquidLiteralBool(True)


def is_whitespace(s: str) -> bool:
    return all(x in ["\t\n "] for x in s)


def flatten_conjunctions(c: Conjunction) -> list[Constraint]:
    queue = [c.c1, c.c2]
    conjunctions = []

    while queue:
        o = queue.pop()
        if isinstance(o, Conjunction):
            queue.append(o.c1)
            queue.append(o.c2)
        elif simplify_is_true(o):
            pass
        else:
            conjunctions.append(o)
    return conjunctions


def is_used_liquid(n: Name, c: LiquidTerm) -> bool:
    return n in liquid_free_vars(c)


def is_used(n: Name, c: Constraint) -> bool:
    if isinstance(c, LiquidConstraint):
        return is_used_liquid(n, c.expr)
    elif isinstance(c, UninterpretedFunctionDeclaration):
        return False
    elif isinstance(c, Implication):
        if n == c.name:
            return False
        return is_used_liquid(n, c.pred) or is_used(n, c.seq)
    elif isinstance(c, Conjunction):
        return is_used(n, c.c1) or is_used(n, c.c2)
    else:
        assert False, f"Unsupported Constraint: {c}"


def simplify_expr(expr: LiquidTerm) -> LiquidTerm:
    """Simplifies a liquid term by reducing it."""
    if isinstance(expr, LiquidApp) and expr.fun == Name("&&", 0):
        if expr.args[0] == LiquidLiteralBool(True):
            return simplify_expr(expr.args[1])
        elif expr.args[1] == LiquidLiteralBool(True):
            return simplify_expr(expr.args[0])
    if isinstance(expr, LiquidApp) and expr.fun == Name("||", 0):
        if expr.args[0] == LiquidLiteralBool(False):
            return simplify_expr(expr.args[1])
        elif expr.args[1] == LiquidLiteralBool(False):
            return simplify_expr(expr.args[0])
    if isinstance(expr, LiquidApp):
        return LiquidApp(expr.fun, [simplify_expr(e) for e in expr.args])
    return expr


def constraint_free_variables(c: Constraint) -> list[Name]:
    """Returns all free variables in a constraint."""
    if isinstance(c, LiquidConstraint):
        return liquid_free_vars(c.expr)
    elif isinstance(c, UninterpretedFunctionDeclaration):
        return []
    elif isinstance(c, Implication):
        lv = liquid_free_vars(c.pred)
        rv = constraint_free_variables(c.seq)
        return [x for x in lv + rv if x != c.name]
    elif isinstance(c, Conjunction):
        return constraint_free_variables(c.c1) + constraint_free_variables(c.c2)
    else:
        assert False, f"Unsupported Constraint: {c}"


def substitution_in_constraint(c: Constraint, rep: LiquidTerm, name: Name) -> Constraint:
    """Substitues a LiquidVar by another expression within a constraint."""
    match c:
        case LiquidConstraint(expr):
            return LiquidConstraint(substitution_in_liquid(expr, rep, name))
        case Conjunction(c1, c2):
            left = substitution_in_constraint(c1, rep, name)
            right = substitution_in_constraint(c2, rep, name)
            return Conjunction(left, right)
        case Implication(name, base, pred, seq):
            if c.name == name:
                return c
            else:
                nseq = substitution_in_constraint(seq, rep, name)
                return Implication(name, base, substitution_in_liquid(pred, rep, name), nseq)
        case UninterpretedFunctionDeclaration(name, type, seq):
            nseq = substitution_in_constraint(seq, rep, name)
            return UninterpretedFunctionDeclaration(name, type, nseq)
        case _:
            assert False


def used_variables(c: LiquidTerm) -> set[Name]:
    """Returns all non-function variables used in an expression."""
    return {x for x in liquid_free_vars(c) if x.name not in base_functions}


def simplify_constraint(c: Constraint) -> Constraint:
    """Converts a constraint into an equivalent one, by reducing it to
    equivalent expressions."""
    if isinstance(c, LiquidConstraint):
        return LiquidConstraint(simplify_expr(c.expr))
    elif isinstance(c, Conjunction):
        left = simplify_constraint(c.c1)
        right = simplify_constraint(c.c2)
        if isinstance(left, LiquidConstraint) and left.expr == LiquidLiteralBool(True):
            return right
        elif isinstance(right, LiquidConstraint) and right.expr == LiquidLiteralBool(True):
            return left
        else:
            return Conjunction(left, right)
    elif isinstance(c, Implication):
        if c.pred == LiquidLiteralBool(True) and c.seq == LiquidConstraint(LiquidLiteralBool(True)):
            return c.seq

        # Preds are usually built as in (cond) && ( this = other)
        if (
            isinstance(c.pred, LiquidApp)
            and c.pred.fun == Name("&&", 0)
            and isinstance(c.pred.args[1], LiquidApp)
            and c.pred.args[1].fun == Name("==", 0)
            and c.pred.args[1].args[0] == LiquidVar(c.name)
        ):
            rep = c.pred.args[1].args[1]
            subs_pred = substitution_in_liquid(c.pred.args[0], rep, c.name)
            subs_seq = substitution_in_constraint(c.seq, rep, c.name)
            rc = simplify_constraint(Implication(Name("_", fresh_counter.fresh()), t_bool, subs_pred, subs_seq))
            return rc

        cont = simplify_constraint(c.seq)
        s = simplify_expr(c.pred)

        other_used_vars = [x for x in used_variables(s) if x != c.name]
        if not is_used(c.name, cont) and not other_used_vars:
            return c.seq

        return Implication(c.name, c.base, s, cont)
    elif isinstance(c, UninterpretedFunctionDeclaration):
        cont = simplify_constraint(c.seq)
        return UninterpretedFunctionDeclaration(c.name, c.type, cont)
    return c


def conjunctive_normal_form(c: Constraint) -> Generator[Constraint, None, None]:
    """Converts a constraint to its conjunctive normal form."""
    if isinstance(c, LiquidConstraint):
        yield c
    elif isinstance(c, Conjunction):
        yield from conjunctive_normal_form(c.c1)
        yield from conjunctive_normal_form(c.c2)
    elif isinstance(c, Implication):
        for inner in conjunctive_normal_form(c.seq):
            yield Implication(c.name, c.base, c.pred, inner)

    elif isinstance(c, UninterpretedFunctionDeclaration):
        for inner in conjunctive_normal_form(c.seq):
            yield UninterpretedFunctionDeclaration(c.name, c.type, inner)
    else:
        assert False


def pretty_print_generator(c: Constraint) -> Generator[tuple[str, int], None, None]:
    """Recursive generates a list of items to print, with the respective
    indentation level."""
    if isinstance(c, LiquidConstraint):
        yield (f"{c.expr}", 0)
    elif isinstance(c, UninterpretedFunctionDeclaration):
        yield (f"fun {c.name} : {c.type}", 0)
        yield from pretty_print_generator(c.seq)
    elif isinstance(c, Implication):
        if is_used(c.name, c.seq):
            yield (f"∀{c.name}:{c.base} | {c.pred}", 0)
        else:
            if c.pred != LiquidLiteralBool(True):
                yield (f"{c.name}:_ | {c.pred}", 0)
        if not isinstance(c.seq, Implication):
            yield ("====>", 0)
        yield from pretty_print_generator(c.seq)
    elif isinstance(c, Conjunction):
        assert False
    else:
        assert False


def is_implication_true(c: Constraint):
    """Returns whether a given constraint has the shape ...

    -> true
    """
    if isinstance(c, LiquidConstraint):
        return c.expr == LiquidLiteralBool(True)
    elif isinstance(c, UninterpretedFunctionDeclaration):
        return is_implication_true(c.seq)
    elif isinstance(c, Implication):
        return is_implication_true(c.seq)
    elif isinstance(c, Conjunction):
        return is_implication_true(c.c1) and is_implication_true(c.c2)
    else:
        assert False


def remove_unrelated_context(c: Constraint, ignore_vars: set[Name]) -> tuple[Constraint, set[Name]]:
    """Removes variables and conditions that are unrelated to the goal."""
    if isinstance(c, LiquidConstraint):
        return (c, used_variables(c.expr).difference(ignore_vars or []))
    elif isinstance(c, UninterpretedFunctionDeclaration):
        return remove_unrelated_context(c.seq, ignore_vars.union([c.name]))
    elif isinstance(c, Implication):
        (ic, vs) = remove_unrelated_context(c.seq, ignore_vars)
        current_vars = used_variables(c.pred).difference(ignore_vars)
        current_vars.add(c.name)
        if vs.isdisjoint(current_vars):
            return (ic, vs)
        else:
            return (c, vs.union(current_vars).difference({c.name}))
    elif isinstance(c, Conjunction):
        (p1, vs1) = remove_unrelated_context(c.c1, ignore_vars)
        (p2, vs2) = remove_unrelated_context(c.c2, ignore_vars)
        return (c, vs1.union(vs2))
    else:
        assert False


def reduce_to_useful_constraint(c: Constraint) -> Constraint:
    top = []
    for cons in conjunctive_normal_form(c):
        if not is_implication_true(cons):
            cons_simp = simplify_constraint(cons)
            cons_clean, _ = remove_unrelated_context(cons_simp, ignore_vars=set())
            top.append(cons_clean)
    llb = LiquidLiteralBool(True)
    return reduce(Conjunction, top, LiquidConstraint(llb))  # type: ignore


def pretty_print_constraint(c: Constraint) -> str:
    """Returns a string representation of a given Constraint.

    To help in reading the constraint, the following optimizations are performed:
    * Conjunctions are expanded into Conjunctive Normal Form.
    * Constrains in the form "sth -> true" are ommited.
    * Constraints are simplified, removing unused variables with no conditions.
    * Constraints are simplified, by removing variables that are not used inside.
    """

    top = []
    for cons in conjunctive_normal_form(c):
        if not is_implication_true(cons):
            r = []
            cons_simp = simplify_constraint(cons)
            cons_clean, _ = remove_unrelated_context(cons_simp, ignore_vars=set())
            for item, indent in pretty_print_generator(cons_clean):
                r.append(indent * "\t" + item)
            top.append("\n".join(r))
    header = "\n+-----Constraint-----+\n"
    middle = "\n+--------------------+\n".join(top)
    footer = "\n+---------//---------+\n"
    return header + middle + footer


def show_constraint(c: Constraint):
    print("Could not show constrain:")
    print(pretty_print_constraint(c))



# This module contain modified code from zc.buildout http://www.buildout.org/en/latest/
# which has this copyright
##############################################################################
#
# Copyright (c) 2005-2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import io
import logging
import re
import urllib.parse
from dataclasses import dataclass
from typing import Dict, List, TextIO

import requests
from lark.exceptions import UnexpectedToken
from lsprotocol.types import Diagnostic, DiagnosticSeverity
from lsprotocol.types import Position, Range

from aeon.facade.driver import AeonDriver
from aeon.lsp.server import AeonLanguageServer

requests_session = requests.Session()

logger = logging.getLogger(__name__)

URI = str
HTTP_SCHEMES = {"http", "https"}


@dataclass(frozen=True)
class ParseResult:
    diagnostics: List[Diagnostic]
    # Not needed for now but could be added on the future
    # core_ast: Any = None
    # typing_ctx: Any = None


_parse_result_cache: Dict[URI, ParseResult] = {}


def clear_cache(uri: URI) -> None:
    """Clear all caches for uri.
    This is to be called when the document is modified.
    """
    logger.debug("Clearing cache for %s", uri)
    _parse_result_cache.pop(uri, None)
    logger.debug("DONE!")


_isurl = re.compile("([a-zA-Z0-9+.-]+)://").match


async def parse(
    aeon_lsp: AeonLanguageServer,
    uri: URI,
    allow_errors: bool = True,
) -> ParseResult:
    if uri in _parse_result_cache:
        return _parse_result_cache[uri]

    parsed_uri = urllib.parse.urlparse(uri)
    if parsed_uri.scheme in HTTP_SCHEMES:
        try:
            response = requests_session.get(uri)
            response.raise_for_status()
            fp = io.StringIO(response.text)
        except requests.RequestException:
            fp = io.StringIO("")
    else:
        document = aeon_lsp.workspace.get_text_document(uri)
        try:
            fp = io.StringIO(document.source)
        except IOError:
            if not allow_errors:
                raise
            fp = io.StringIO("")

    parse_result = await _parse(fp, aeon_lsp.aeon_driver)
    _parse_result_cache[uri] = parse_result
    return parse_result


async def _parse(
    fp: TextIO,
    driver: AeonDriver,
) -> ParseResult:
    diagnostics = []

    try:
        content = fp.read()
        errors = driver.parse(aeon_code=content)

        for error in errors:
            error_message = str(error)
            error_position = error.position()

            error_start_line, error_start_character = error_position.get_start()
            error_end_line, error_end_character = error_position.get_end()

            error_range = Range(
                start=Position(line=error_start_line, character=error_start_character),
                end=Position(line=error_end_line, character=error_end_character),
            )

            diagnostics.append(
                Diagnostic(
                    message=error_message,
                    range=error_range,
                    source="aeon",
                    severity=DiagnosticSeverity.Error,
                )
            )

    except UnexpectedToken as e:
        token_length = 1
        try:
            token_length = len(str(e.token.value))
        except Exception:
            pass

        token_type = e.token.type if e.token.type else "unknown"
        token_value = e.token.value if e.token.value else str(e.token)

        error_message = (
            f"Unexpected {token_type} '{token_value}' at line {e.line}, column {e.column}.\n"
            f"Expected: {', '.join(e.expected)}"
        )

        error_range = Range(
            start=Position(line=e.line - 1, character=e.column - 1),
            end=Position(line=e.line - 1, character=e.column - 1 + token_length),
        )

        diagnostics.append(
            Diagnostic(
                message=error_message,
                range=error_range,
                source="aeon",
                severity=DiagnosticSeverity.Error,
            )
        )
    except Exception as e:
        logger.exception("Unhandled exception while parsing")
        diagnostics.append(
            Diagnostic(
                message=f"Unknown exception '{e}' occurred while parsing",
                range=Range(
                    start=Position(line=0, character=0),
                    end=Position(line=0, character=0),
                ),
                source="aeon",
                severity=DiagnosticSeverity.Error,
            )
        )
    return ParseResult(diagnostics)


async def _open(
    aeon_lsp: AeonLanguageServer,
    base: str,
    uri: URI,
    allow_errors: bool,
) -> ParseResult:
    logger.debug("Opening URI for parsing: base=%r, uri=%r", base, uri)
    if not _isurl(uri):
        assert base
        uri = urllib.parse.urljoin(base, uri)
    parse_result = await parse(aeon_lsp, uri, allow_errors=allow_errors)
    return parse_result



from aeon.utils.name import Name
from aeon.sugar.stypes import (
    SAbstractionType,
    STypeConstructor,
    STypeVar,
    SRefinedType,
    SType,
    STypePolymorphism,
)
from aeon.sugar.program import (
    SAbstraction,
    SAnnotation,
    SApplication,
    SHole,
    SIf,
    SLet,
    SLiteral,
    SRec,
    STerm,
    STypeAbstraction,
    STypeApplication,
    SVar,
)


def normalize(ty: SType) -> SType:
    match ty:
        case SRefinedType(oname, SRefinedType(iname, ity, iref), oref):
            a1 = substitution_sterm_in_sterm(
                iref,
                SVar(oname),
                iname,
            )
            new_ref = SApplication(SApplication(SVar(Name("&&", 0)), a1), oref)
            return SRefinedType(oname, ity, new_ref)
        case _:
            return ty


def substitute_svartype_in_stype(ty: SType, beta: SType, alpha: Name):
    """Replaces all occurrences of vartypes name in t by rep."""

    def rec(k: SType):
        return substitute_svartype_in_stype(k, beta, alpha)

    ty = normalize(ty)
    match ty:
        case STypeVar(tname):
            if tname == alpha:
                return beta
            else:
                return ty
        case SRefinedType(name, ty, ref):
            return SRefinedType(name, rec(ty), ref)
        case SAbstractionType(var_name, var_type, return_type):
            return SAbstractionType(var_name, rec(var_type), rec(return_type))
        case STypePolymorphism(tname, kind, body):
            if tname == alpha:
                return ty
            else:
                return STypePolymorphism(tname, kind, rec(body))
        case STypeConstructor(name, args):
            return STypeConstructor(name, [rec(a) for a in args])
        case _:
            assert False, f"Unknown node in substitute {ty}"


def substitution_sterm_in_stype(ty: SType, beta: STerm, alpha: Name) -> SType:
    def rec(k: SType):
        return substitution_sterm_in_stype(k, beta, alpha)

    match ty:
        case STypeVar(_):
            return ty
        case SRefinedType(name, ty, ref):
            return SRefinedType(name, rec(ty), substitution_sterm_in_sterm(ref, beta, alpha))
        case SAbstractionType(var_name, var_type, return_type):
            return SAbstractionType(var_name, rec(var_type), rec(return_type))
        case STypePolymorphism(tname, kind, body):
            return STypePolymorphism(tname, kind, rec(body))
        case STypeConstructor(name, args):
            return STypeConstructor(name, [rec(a) for a in args])
        case _:
            assert False


def substitution_sterm_in_sterm(t: STerm, beta: STerm, alpha: Name) -> STerm:
    def rec(x: STerm):
        return substitution_sterm_in_sterm(x, beta, alpha)

    def rect(x: SType):
        return substitution_sterm_in_stype(x, beta, alpha)

    match t:
        case SLiteral(_, _) | SHole(_):
            return t
        case SVar(name):
            if name == alpha:
                return beta
            else:
                return t
        case SApplication(fun, arg, loc):
            return SApplication(rec(fun), rec(arg), loc=loc)
        case SAbstraction(aname, body, loc):
            if aname == alpha:
                return t
            else:
                return SAbstraction(aname, rec(body), loc=loc)
        case SLet(vname, vvalue, body, loc):
            if vname == alpha:
                return SLet(vname, rec(vvalue), body, loc=loc)
            else:
                return SLet(vname, rec(vvalue), rec(body), loc=loc)
        case SRec(vname, vty, vvalue, body, loc):
            if vname == alpha:
                return SRec(vname, rect(vty), rec(vvalue), body, loc=loc)
            else:
                return SRec(vname, rect(vty), rec(vvalue), rec(body), loc=loc)
        case SAnnotation(expr, ty, loc):
            return SAnnotation(rec(expr), rect(ty), loc=loc)
        case SIf(cond, then, otherwise, loc):
            return SIf(rec(cond), rec(then), rec(otherwise), loc=loc)
        case STypeApplication(body, ty, loc):
            return STypeApplication(rec(body), rect(ty), loc=loc)
        case STypeAbstraction(aname, kind, body, loc):
            return STypeAbstraction(aname, kind, rec(body), loc=loc)
        case _:
            assert False


def substitution_svartype_in_sterm(t: STerm, rep: SType, name: Name) -> STerm:
    def rec(x: STerm):
        return substitution_svartype_in_sterm(x, rep, name)

    match t:
        case SVar(_) | SHole(_):
            return t
        case SLiteral(v, ty, loc):
            return SLiteral(v, substitute_svartype_in_stype(ty, rep, name), loc=loc)
        case SApplication(fun, arg, loc):
            return SApplication(rec(fun), rec(arg), loc=loc)
        case SAbstraction(aname, body, loc):
            return SAbstraction(aname, rec(body), loc=loc)
        case SLet(vname, vvalue, body, loc):
            return SLet(vname, rec(vvalue), rec(body), loc=loc)
        case SRec(vname, vtype, vvalue, body, loc):
            return SRec(vname, substitute_svartype_in_stype(vtype, rep, name), rec(vvalue), rec(body), loc=loc)
        case SAnnotation(expr, ty, loc):
            return SAnnotation(rec(expr), substitute_svartype_in_stype(ty, rep, name), loc=loc)
        case SIf(cond, then, otherwise, loc):
            return SIf(rec(cond), rec(then), rec(otherwise), loc=loc)
        case STypeApplication(body, ty, loc):
            return STypeApplication(rec(body), substitute_svartype_in_stype(ty, rep, name), loc=loc)
        case STypeAbstraction(aname, kind, body, loc):
            if aname == name:
                return t
            else:
                return STypeAbstraction(aname, kind, rec(body), loc=loc)
        case _:
            assert False



from aeon.core.terms import (
    Abstraction,
    Annotation,
    Application,
    If,
    Let,
    Literal,
    Rec,
    Term,
    TypeAbstraction,
    TypeApplication,
    Var,
)

from aeon.utils.name import Name, fresh_counter


class ANFConverter:
    """Recursive visitor that applies ANF transformation."""

    def fresh(self) -> Name:
        return Name("anf", fresh_counter.fresh())

    def convert(self, t: Term):
        """Converts term to ANF form."""

        match t:
            case If(cond=cond, then=then, otherwise=otherwise, loc=loc):
                cond = self.convert(cond)
                then = self.convert(then)
                otherwise = self.convert(otherwise)
                if isinstance(cond, Var) or isinstance(cond, Literal):
                    return If(cond, then, otherwise, loc=loc)
                else:
                    v = self.fresh()
                    return self.convert(Let(v, cond, If(Var(v, cond.loc), then, otherwise, loc=loc), loc=cond.loc))
            case Application(fun=fun, arg=arg, loc=loc):
                fun = self.convert(fun)

                if isinstance(fun, Var) or isinstance(fun, Literal):
                    pass
                elif isinstance(fun, Let):
                    return Let(fun.var_name, fun.var_value, self.convert(Application(fun.body, arg, loc=loc)), loc=loc)
                else:
                    v = self.fresh()
                    return self.convert(Let(v, fun, Application(Var(v), arg, loc=loc), loc=loc))

                arg = self.convert(arg)
                if isinstance(arg, Var) or isinstance(arg, Literal):
                    return Application(fun, arg, loc=loc)
                else:
                    v = self.fresh()
                    return self.convert(Let(v, arg, Application(fun, Var(v), loc=loc), loc=loc))

            case Let(var_name=name, var_value=value, body=body, loc=loc):
                value = self.convert(value)
                body = self.convert(body)
                match value:
                    case Let(var_name=vname, var_value=vvalue, body=vbody, loc=iloc):
                        assert name != vname
                        vvalue = self.convert(vvalue)
                        vbody = self.convert(vbody)
                        return Let(vname, vvalue, self.convert(Let(name, vbody, body, loc=loc)), loc=iloc)
                    case Rec(var_name=vname, var_type=vtype, var_value=vvalue, body=vbody, loc=loc):
                        assert name != vname
                        vvalue = self.convert(vvalue)
                        vbody = self.convert(vbody)
                        return Rec(vname, vtype, vvalue, self.convert(Let(name, vbody, body, loc=loc)), loc=loc)
                    case _:
                        return Let(name, value, body, loc=loc)
            case Rec(var_name=name, var_type=type, var_value=value, body=body, loc=loc):
                value = self.convert(value)
                body = self.convert(body)
                return Rec(name, type, value, body, loc=loc)
            case Abstraction(var_name=name, body=body, loc=loc):
                body = self.convert(body)
                return Abstraction(var_name=name, body=body, loc=loc)
            case Annotation(expr=expr, type=ty, loc=loc):
                expr = self.convert(expr)
                return Annotation(expr=expr, type=ty, loc=loc)
            case TypeAbstraction(name=name, kind=kind, body=body, loc=loc):
                body = self.convert(body)
                return TypeAbstraction(name, kind, body, loc=loc)
            case TypeApplication(body=body, type=type, loc=loc):
                body = self.convert(body)
                return TypeApplication(body, type, loc=loc)
            case _:
                return t


def ensure_anf(t: Term) -> Term:
    """Converts a term to ANF form."""

    return ANFConverter().convert(t)



from __future__ import annotations

from functools import reduce
from typing import Any, Generator

from aeon.core.types import builtin_core_types
from aeon.core.liquid import LiquidApp
from aeon.core.types import LiquidHornApplication, TypeConstructor
from aeon.core.liquid import LiquidLiteralBool
from aeon.core.liquid import LiquidLiteralFloat
from aeon.core.liquid import LiquidLiteralInt
from aeon.core.liquid import LiquidLiteralString
from aeon.core.liquid import LiquidTerm
from aeon.core.liquid import LiquidVar
from aeon.core.liquid_ops import mk_liquid_and
from aeon.core.substitutions import substitution_in_liquid
from aeon.core.types import AbstractionType
from aeon.core.types import RefinedType
from aeon.core.types import t_bool
from aeon.core.types import Top
from aeon.core.types import Type
from aeon.core.types import TypePolymorphism
from aeon.core.types import TypeVar
from aeon.core.liquid_ops import liquid_prelude
from aeon.typechecking.context import TypingContext
from aeon.typechecking.liquid import LiquidTypeCheckingContext, check_liquid
from aeon.verification.helpers import constraint_builder
from aeon.verification.helpers import end
from aeon.verification.helpers import imp
from aeon.verification.smt import smt_valid
from aeon.verification.vcs import Conjunction
from aeon.verification.vcs import Constraint
from aeon.verification.vcs import Implication
from aeon.verification.vcs import LiquidConstraint
from aeon.verification.vcs import UninterpretedFunctionDeclaration
from aeon.utils.name import Name, fresh_counter

Assignment = dict[Name, list[LiquidTerm]]


def smt_base_type(ty: Type) -> str | None:
    if isinstance(ty, AbstractionType):
        return None
    if isinstance(ty, TypeConstructor):
        return str(ty)
    elif isinstance(ty, RefinedType):
        return smt_base_type(ty.type)
    else:
        return None


def fresh(context: TypingContext, ty: Type) -> Type:
    match ty:
        case RefinedType(name, ity, LiquidHornApplication(_, _)):
            vname = Name("√", fresh_counter.fresh())
            hole_name = Name("hole", fresh_counter.fresh())

            # TODO Poly: check if t should be in LiquidTypes
            return RefinedType(
                vname,
                ity,
                LiquidHornApplication(
                    hole_name,
                    [
                        (LiquidVar(n), t)
                        for n, t in context.vars() + [(vname, ty.type)]
                        if isinstance(t, TypeConstructor)
                    ],
                ),
            )
        case RefinedType(_, _, _) | Top() | TypeVar():
            return ty
        case AbstractionType(name, aty, rty):
            sp = fresh(context, aty)
            tp = fresh(context.with_var(name, aty), rty)
            return AbstractionType(name, sp, tp)
        case TypePolymorphism(name, kind, body):
            return TypePolymorphism(name, kind, fresh(context, body))
        case TypeConstructor(name, args):
            return TypeConstructor(name, [fresh(context, c) for c in args])
        case _:
            assert False, f"Type not freshable: {ty}, {type(ty)}"


def obtain_holes(t: LiquidTerm) -> list[LiquidHornApplication]:
    if isinstance(t, LiquidHornApplication):
        return [t]
    elif (
        isinstance(t, LiquidLiteralBool)
        or isinstance(t, LiquidLiteralInt)
        or isinstance(t, LiquidLiteralFloat)
        or isinstance(t, LiquidLiteralString)
    ):
        return []
    elif isinstance(t, LiquidVar):
        return []
    elif isinstance(t, LiquidApp):
        holes: list[LiquidHornApplication] = []
        for h in t.args:
            holes = holes + obtain_holes(h)
        return holes
    else:
        assert False, f"Unkown term type: {t} ({type(t)})"


def obtain_holes_constraint(c: Constraint) -> list[LiquidHornApplication]:
    match c:
        case Conjunction(c1, c2):
            return obtain_holes_constraint(c1) + obtain_holes_constraint(c2)
        case Implication(_, _, pre, post):
            return obtain_holes(pre) + obtain_holes_constraint(post)
        case LiquidConstraint(e):
            return obtain_holes(e)
        case UninterpretedFunctionDeclaration(_, _, post):
            return obtain_holes_constraint(post)
        case _:
            assert False, c


def contains_horn(t: LiquidTerm) -> bool:
    if (
        isinstance(t, LiquidLiteralInt)
        or isinstance(t, LiquidLiteralBool)
        or isinstance(t, LiquidLiteralString)
        or isinstance(t, LiquidLiteralFloat)
    ):
        return False
    elif isinstance(t, LiquidVar):
        return False
    elif isinstance(t, LiquidHornApplication):
        return True
    elif isinstance(t, LiquidApp):
        return all([contains_horn(arg) for arg in t.args])
    else:
        assert False


def contains_horn_constraint(c: Constraint) -> bool:
    if isinstance(c, LiquidConstraint):
        return contains_horn(c.expr)
    elif isinstance(c, Conjunction):
        return contains_horn_constraint(c.c1) or contains_horn_constraint(c.c2)
    elif isinstance(c, Implication):
        return contains_horn(c.pred) or contains_horn_constraint(c.seq)
    elif isinstance(c, UninterpretedFunctionDeclaration):
        return contains_horn_constraint(c.seq)
    else:
        assert False


def wellformed_horn(predicate: LiquidTerm) -> bool:
    if not contains_horn(predicate):
        return True
    elif (
        isinstance(predicate, LiquidApp)
        and predicate.fun == "&&"
        and not contains_horn(predicate.args[0])
        and isinstance(predicate.args[1], LiquidHornApplication)
    ):
        return True
    elif isinstance(predicate, LiquidHornApplication):
        return True
    else:
        return False


def mk_arg(i: int) -> Name:
    return Name(f"arg_{i}", 0)


def get_possible_args(vars: list[tuple[LiquidTerm, TypeConstructor | TypeVar]], arity: int):
    if arity == 0:
        yield []
    else:
        for base in get_possible_args(vars, arity - 1):
            for i, (_, _) in enumerate(vars):
                yield [LiquidVar(mk_arg(i))] + base
                yield [LiquidLiteralBool(True)] + base
                yield [LiquidLiteralBool(False)] + base
                yield [LiquidLiteralInt(0)] + base
                yield [LiquidLiteralInt(1)] + base


def build_possible_assignment(hole: LiquidHornApplication) -> Generator[LiquidApp]:
    ctx = LiquidTypeCheckingContext(
        known_types=[TypeConstructor(Name(bn.name.name, 0)) for bn in builtin_core_types],
        functions=liquid_prelude,
        variables={mk_arg(i): t for i, (_, t) in enumerate(hole.argtypes)},
    )

    for fname in liquid_prelude:
        ftype = liquid_prelude[fname]
        arity = len(ftype) - 1
        for args in get_possible_args(hole.argtypes, arity):
            # At least one LiquidVar must be used.
            if not any(isinstance(a, LiquidVar) for a in args):
                continue
            arg_list = list(args)
            app = LiquidApp(fname, arg_list)
            if check_liquid(ctx, app, t_bool):
                yield app


def build_initial_assignment(c: Constraint) -> Assignment:
    holes = obtain_holes_constraint(c)
    assign: dict[Name, list[LiquidTerm]] = {}
    for h in holes:
        if h.name not in assign:
            assign[h.name] = list(build_possible_assignment(h))
    return assign


def merge_assignments(xs: list[LiquidTerm]) -> LiquidTerm:
    b = LiquidLiteralBool(True)
    for c in xs:
        b = mk_liquid_and(b, c)
    return b


def split(c: Constraint) -> list[Constraint]:
    match c:
        case LiquidConstraint(_):
            return [c]
        case Conjunction(c1, c2):
            return split(c1) + split(c2)
        case Implication(name, base, pre, post):
            return [Implication(name, base, pre, cp) for cp in split(post)]
        case UninterpretedFunctionDeclaration(name, type, seq):
            return [UninterpretedFunctionDeclaration(name, type, c) for c in split(seq)]
        case _:
            assert False


def build_forall_implication(
    vs: list[tuple[Name, Type]],
    p: LiquidTerm,
    c: Constraint,
) -> Constraint:
    if not vs:
        return c
    else:
        for name, _ in vs:
            assert isinstance(name, Name)
    lastEl = vs[-1]
    assert isinstance(lastEl[1], TypeConstructor)
    cf = Implication(lastEl[0], lastEl[1], p, c)
    for n, t in vs[-2::-1]:
        assert isinstance(t, TypeConstructor)
        assert isinstance(n, Name)
        cf = Implication(n, t, LiquidLiteralBool(True), cf)
    return cf


def simpl(vs: list[tuple[Name, Type]], p: LiquidTerm, c: Constraint) -> Constraint:
    if isinstance(c, Implication):
        return simpl(vs + [(c.name, c.base)], mk_liquid_and(p, c.pred), c.seq)
    else:
        return build_forall_implication(vs, p, c)


def flat(c: Constraint) -> list[Constraint]:
    return [simpl([], LiquidLiteralBool(True), cp) for cp in split(c)]


def has_k_head(c: Constraint) -> bool:
    match c:
        case Conjunction(_, _):
            assert False
        case Implication(_, _, _, post):
            return has_k_head(post)
        case LiquidConstraint(e):
            return isinstance(e, LiquidHornApplication)
        case UninterpretedFunctionDeclaration(_, _, post):
            return has_k_head(post)
        case _:
            assert False, f"Unkown constraint type: {c} ({type(c)})"


def apply_constraint(assign: Assignment, c: Constraint) -> Constraint:
    match c:
        case LiquidConstraint(e):
            return LiquidConstraint(apply_liquid(assign, e))
        case Conjunction(c1, c2):
            return Conjunction(
                apply_constraint(assign, c1),
                apply_constraint(assign, c2),
            )
        case Implication(name, base, pre, post):
            return Implication(
                name,
                base,
                apply_liquid(assign, pre),
                apply_constraint(assign, post),
            )
        case UninterpretedFunctionDeclaration(name, base, post):
            return UninterpretedFunctionDeclaration(name, base, apply_constraint(assign, post))
        case _:
            assert False


def fill_horn_arguments(h: LiquidHornApplication, candidate: LiquidTerm) -> LiquidTerm:
    for i, (n, _) in enumerate(h.argtypes):
        assert isinstance(n, LiquidTerm)
        candidate = substitution_in_liquid(candidate, n, mk_arg(i))
    return candidate


def apply_liquid(assign: Assignment, c: LiquidTerm) -> LiquidTerm:
    if isinstance(c, LiquidHornApplication):
        if c.name in assign:
            ne = assign[c.name]
            return fill_horn_arguments(c, merge_assignments(ne))
        else:
            return c
    elif isinstance(c, LiquidApp):
        return LiquidApp(c.fun, [apply_liquid(assign, ci) for ci in c.args])
    else:
        return c


def apply(assign: Assignment, c: Any):
    if isinstance(c, Constraint):
        return apply_constraint(assign, c)
    elif isinstance(c, LiquidTerm):
        return apply_liquid(assign, c)
    assert False


def extract_components_of_imp(
    c: Constraint,
) -> tuple[list[tuple[Name, TypeConstructor | TypeVar | AbstractionType | Top]], tuple[LiquidTerm, LiquidTerm]]:
    match c:
        case UninterpretedFunctionDeclaration(name, base, post):
            (vs1, (p, h)) = extract_components_of_imp(post)
            vsh: list[tuple[Name, TypeConstructor | TypeVar | AbstractionType | Top]] = [(name, base)]
            return (vsh + vs1, (p, h))
        case Implication(name, base, pre, seq):
            (vs1, (p, h)) = extract_components_of_imp(seq)
            vs: list[tuple[Name, TypeConstructor | TypeVar | AbstractionType | Top]] = [(name, base)]
            return (vs + vs1, (mk_liquid_and(pre, p), h))
        case LiquidConstraint(e):
            return ([], (LiquidLiteralBool(True), e))
        case _:
            assert False, f"Unkown context: {c} ({type(c)})"


def weaken(assign, c: Constraint) -> Assignment:
    (vs, (p, h)) = extract_components_of_imp(c)
    assert isinstance(h, LiquidHornApplication)
    assert h.name in assign
    current_rep = assign[h.name]

    def keep(q: LiquidTerm) -> bool:
        assert isinstance(h, LiquidHornApplication)
        qp = fill_horn_arguments(h, q)
        nc = constraint_builder(vs, imp(apply(assign, p), end(qp)))
        return smt_valid(nc)

    qsp = [q for q in current_rep if keep(q)]
    return {k: assign[k] if k != h.name else qsp for k in assign}


def fixpoint(cs: list[Constraint], assign) -> Assignment:
    ncs = [c for c in cs if not smt_valid(apply(assign, c))]
    if not ncs:
        return assign
    else:
        weakened_assignment = weaken(assign, ncs[0])
        return fixpoint(cs, weakened_assignment)


# TODO uninterpreted: We need to pass the context here, to use custom measures in the horn clause.


def solve(c: Constraint) -> bool:
    # Performance improvement
    if not contains_horn_constraint(c):
        # TODO: Try to simplify the expression before sending to the SMT solver
        # v = reduce_to_useful_constraint(c)
        return smt_valid(c)
    cs = flat(c)
    csk = [c for c in cs if has_k_head(c)]
    csp = [c for c in cs if not has_k_head(c)]
    assignment0: Assignment = build_initial_assignment(c)
    subst = fixpoint(csk, assignment0)

    def merge(acc: Constraint, pi: Constraint) -> Constraint:
        return Conjunction(acc, pi)

    merged_csps: Constraint
    seed: Constraint = LiquidConstraint(LiquidLiteralBool(True))
    merged_csps = reduce(merge, csp, seed)
    c_final: Constraint = apply(subst, merged_csps)
    return smt_valid(c_final)



from aeon.core.liquid import (
    LiquidApp,
    LiquidLiteralBool,
    LiquidLiteralFloat,
    LiquidLiteralInt,
    LiquidLiteralString,
    LiquidTerm,
    LiquidVar,
)
from aeon.core.terms import (
    Term,
    Literal,
    Var,
    Annotation,
    Hole,
    Application,
    Abstraction,
    Let,
    Rec,
    If,
    TypeAbstraction,
    TypeApplication,
)
from aeon.core.types import AbstractionType, RefinedType, Top, Type, TypeConstructor, TypePolymorphism, TypeVar
from aeon.sugar.program import (
    STerm,
    SLiteral,
    SVar,
    SAnnotation,
    SHole,
    SApplication,
    SAbstraction,
    SLet,
    SRec,
    SIf,
    STypeAbstraction,
    STypeApplication,
)
from aeon.sugar.stypes import SAbstractionType, SRefinedType, SType, STypeConstructor, STypePolymorphism, STypeVar
from aeon.sugar.ast_helpers import st_bool, st_int, st_float, st_string, st_top


def lift_liquid(ref: LiquidTerm) -> STerm:
    match ref:
        case LiquidLiteralBool(value, loc):
            return SLiteral(value, st_bool, loc)
        case LiquidLiteralInt(value, loc):
            return SLiteral(value, st_int, loc)
        case LiquidLiteralFloat(value, loc):
            return SLiteral(value, st_float, loc)
        case LiquidLiteralString(value, loc):
            return SLiteral(value, st_string, loc)
        case LiquidVar(name, loc):
            return SVar(name, loc)
        case LiquidApp(fun, args, loc):
            v: STerm = SVar(fun, loc=loc)
            for arg in args[::-1]:
                v = SApplication(v, lift_liquid(arg), loc=loc)
            return v
        case _:
            assert False, f"Don't know how to lift liquid term {ref} ({type(ref)})"


def lift_type(ty: Type) -> SType:
    match ty:
        case Top():
            return st_top
        case TypeVar(name, loc):
            return STypeVar(name, loc=loc)
        case TypeConstructor(name, args, loc):
            return STypeConstructor(name, [lift_type(arg) for arg in args], loc=loc)
        case AbstractionType(name, atype, rtype, loc):
            return SAbstractionType(name, lift_type(atype), lift_type(rtype), loc=loc)
        case RefinedType(name, typ, ref, loc):
            return SRefinedType(name, lift_type(typ), lift_liquid(ref), loc=loc)
        case TypePolymorphism(name, kind, body, loc):
            return STypePolymorphism(name, kind, lift_type(body), loc=loc)
        case _:
            assert False, f"Don't know how to lift type {ty} ({type(ty)})"


def lift(t: Term) -> STerm:
    match t:
        case Literal(value, typ, loc):
            # Convert core type to surface type (simple wrapper)
            return SLiteral(value, lift_type(typ), loc=loc)
        case Var(name, loc):
            return SVar(name, loc=loc)
        case Annotation(expr, typ, loc):
            return SAnnotation(lift(expr), lift_type(typ), loc=loc)
        case Hole(name, loc):
            return SHole(name, loc=loc)
        case Application(fun, arg, loc):
            return SApplication(lift(fun), lift(arg), loc=loc)
        case Abstraction(var_name, body, loc):
            return SAbstraction(var_name, lift(body), loc=loc)
        case Let(var_name, var_value, body, loc):
            return SLet(var_name, lift(var_value), lift(body), loc=loc)
        case Rec(var_name, var_type, var_value, body, loc):
            return SRec(var_name, lift_type(var_type), lift(var_value), lift(body), loc=loc)
        case If(cond, then, otherwise, loc):
            return SIf(lift(cond), lift(then), lift(otherwise), loc=loc)
        case TypeAbstraction(name, kind, body, loc):
            return STypeAbstraction(name, kind, lift(body), loc=loc)
        case TypeApplication(body, typ, loc):
            return STypeApplication(lift(body), lift_type(typ), loc=loc)
        case _:
            assert False, f"Don't know how to lift {t} ({type(t)})"



from __future__ import annotations

import sys
import time
from typing import Any, Optional
from typing import Callable
from typing import TypeAlias

import multiprocess as mp
from loguru import logger

from aeon.backend.evaluator import EvaluationContext
from aeon.core.substitutions import substitution
from aeon.core.terms import Term, Var
from aeon.core.types import Top
from aeon.core.types import top, Type
from aeon.decorators import Metadata
from aeon.frontend.anf_converter import ensure_anf
from aeon.backend.evaluator import eval
from aeon.synthesis.uis.api import SynthesisUI
from aeon.synthesis.identification import get_holes_info
from aeon.typechecking.context import TypingContext
from aeon.typechecking.typeinfer import check_type
from aeon.utils.name import Name

from aeon.synthesis.api import ErrorInSynthesis, Synthesizer, TimeoutInEvaluationException

from aeon.synthesis.decorators import Goal


def make_program(whole_program: Term, hole_name: Name) -> Callable[[Term], Term]:
    def replace(candidate: Term) -> Term:
        new_program = substitution(whole_program, candidate, hole_name)
        core_ast_anf = ensure_anf(new_program)
        return core_ast_anf

    return replace


def make_validator(ctx: TypingContext, replace: Callable[[Term], Term]) -> Callable[[Term], bool]:
    def validate(candidate: Term) -> bool:
        prog = replace(candidate)
        return check_type(ctx, prog, Top())

    return validate


Evaluators: TypeAlias = list[Callable[[Term], float]]


def make_evaluators(ectx: EvaluationContext, fun_name: Name, metadata: Metadata) -> Evaluators:
    """Returns a list of functions that take the original program and return each fitness value"""

    goals: list[Goal] = metadata.get(fun_name, {}).get("goals", [])
    fitnesses: list[Callable[[Term], float]] = []
    for goal in goals:
        assert goal.length == 1, "Currently, we only support 1 fitness value per function"

        def fitness(v: Term) -> float:
            program_for_fitness = substitution(v, Var(goal.function), Name("main", 0))
            try:
                result = eval(program_for_fitness, ectx)
            except Exception:
                result = sys.maxsize
            return result

        fitnesses.append(fitness)
    return fitnesses


def make_evaluator(
    ectx: EvaluationContext, replace: Callable[[Term], Term], evaluators: Evaluators, budget_eval: float = 1.0
) -> Callable[[Term], list[float]]:
    """Creates a function that takes candidate programs and return the fitness list."""

    def evaluate_individual(program: Any, result_queue: mp.Queue) -> Any:
        """Function to run in a separate process and places the result in a Queue."""
        start = time.time()
        try:
            results = [ev(program) for ev in evaluators]
            assert isinstance(results, list)
            result_queue.put(results)
        except Exception as e:
            logger.log("SYNTHESIZER", f"Failed in the fitness function: {e}, {type(e)}")
            raise ErrorInSynthesis(e, msg=f"Failed in the fitness function: {e}, {type(e)}")
        finally:
            end = time.time()
            logger.info(f"Individual evaluation time: {end - start} ")

    def evaluate(candidate: Term, timeout: float = budget_eval) -> list[float]:
        # import faulthandler; faulthandler.enable()
        prog = replace(candidate)
        result_queue = mp.Queue()
        eval_process = mp.Process(target=evaluate_individual, args=(prog, result_queue))
        eval_process.start()
        eval_process.join(timeout=timeout)
        if eval_process.is_alive():
            eval_process.close()
            eval_process.terminate()
            eval_process.join()
            raise TimeoutInEvaluationException()
        else:
            fitness_values = result_queue.get()
            return fitness_values

    return evaluate


def synthesize_holes(
    ctx: TypingContext,
    ectx: EvaluationContext,
    term: Term,
    targets: list[tuple[Name, list[Name]]],
    metadata: Metadata,
    synthesizer: Synthesizer,
    budget: float = 60.0,
    ui: SynthesisUI = SynthesisUI(),
    budget_eval: Optional[float] = None,
) -> dict[Name, Optional[Term]]:
    """Synthesizes code for multiple functions, each with multiple holes."""

    if budget_eval is None:
        budget_eval = max(budget / 1000, 1)

    program_holes = get_holes_info(ctx, term, top, targets)

    mapping = {}

    for fun_name, holes_names in targets:
        assert len(holes_names) == 1, "Currently, we only support 1 hole per function"

        hole_name = holes_names[0]
        ty, tyctx = program_holes[hole_name]
        ui.start(tyctx, ectx, hole_name.name, ty, budget)

        replace = make_program(term, hole_name)
        validator = make_validator(ctx, replace)
        evaluators = make_evaluators(ectx, fun_name, metadata)
        evaluator = make_evaluator(ectx, replace, evaluators, budget_eval)
        assert isinstance(tyctx, TypingContext)
        assert isinstance(ty, Type)
        try:
            t = synthesizer.synthesize(
                ctx=tyctx,
                type=ty,
                validate=validator,
                evaluate=evaluator,
                fun_name=fun_name,
                metadata=metadata,
                budget=budget,
                ui=ui,
            )
        except Exception as e:
            ui.end(None, None)
            raise e

        ui.end(t, None)
        mapping[hole_name] = t
    return mapping



import asyncio
from typing import List, Optional, AsyncIterable

from lsprotocol.types import (
    TEXT_DOCUMENT_COMPLETION,
    TEXT_DOCUMENT_DID_OPEN,
    TEXT_DOCUMENT_DID_CHANGE,
    WORKSPACE_DID_CHANGE_WATCHED_FILES,
    CompletionItem,
    CompletionOptions,
    CompletionParams,
    DidChangeTextDocumentParams,
    DidChangeWatchedFilesParams,
    DidOpenTextDocumentParams,
    Diagnostic,
)
from pygls.server import LanguageServer

from aeon.facade.driver import AeonDriver


class AeonLanguageServer(LanguageServer):
    def __init__(self, aeon_driver: AeonDriver):
        super().__init__("aeon.lsp.server", "0.1.0")
        self.aeon_driver = aeon_driver
        self.debounce_delay = 0.3
        self._setup_handlers()

    def start(self, tcp_server):
        if not tcp_server:
            self.start_io()

        host, port = tcp_server.split(":") if ":" in tcp_server else ("localhost", tcp_server)

        print(f"Listening on {host}:{port}")
        self.start_tcp(host, int(port))

    async def _parse_and_send_diagnostics(self, uri) -> None:
        await asyncio.sleep(self.debounce_delay)
        diagnostics = []
        async for diag in self._get_diagnostics(uri):
            diagnostics.append(diag)
        self.publish_diagnostics(uri, diagnostics)

    async def _get_diagnostics(self, uri: str) -> AsyncIterable[Diagnostic]:
        from . import aeon_adapter

        ast = await aeon_adapter.parse(self, uri)
        for diag in ast.diagnostics:
            yield diag

    def _setup_handlers(self):
        @self.feature(TEXT_DOCUMENT_DID_OPEN)
        async def did_open(
            ls: AeonLanguageServer,
            params: DidOpenTextDocumentParams,
        ) -> None:
            await ls._parse_and_send_diagnostics(params.text_document.uri)

        @self.feature(TEXT_DOCUMENT_DID_CHANGE)
        async def did_change(
            ls: AeonLanguageServer,
            params: DidChangeTextDocumentParams,
        ) -> None:
            from . import aeon_adapter

            aeon_adapter.clear_cache(params.text_document.uri)
            await ls._parse_and_send_diagnostics(params.text_document.uri)

        @self.feature(WORKSPACE_DID_CHANGE_WATCHED_FILES)
        async def did_change_watched_file(
            ls: AeonLanguageServer,
            params: DidChangeWatchedFilesParams,
        ) -> None:
            from . import aeon_adapter

            for change in params.changes:
                aeon_adapter.clear_cache(change.uri)

        @self.feature(TEXT_DOCUMENT_COMPLETION, CompletionOptions(trigger_characters=["= "]))
        async def lsp_completion(
            ls: AeonLanguageServer,
            params: CompletionParams,
        ) -> Optional[List[CompletionItem]]:
            await asyncio.sleep(self.debounce_delay)
            return []  # TODO
            # items: List[CompletionItem] = []
            #
            # ast = await buildout.parse(ls, params.text_document.uri, True)
            # for line in ast.lines:
            #   pos = params.position
            #   (var_name, var_type, value) = line
            #   ci = CompletionItem(
            #       label=var_name,
            #       text_edit=TextEdit(
            #           range=Range(start=Position(line=pos.line, character=pos.character),
            #                       end=Position(line=pos.line,
            #                                    character=pos.character + len(var_name))),
            #           new_text=var_name,
            #       ),
            #       kind=CompletionItemKind.Variable,
            #       documentation=MarkupContent(
            #           kind=MarkupKind.Markdown,
            #           value=f"{var_name} : {var_type} = {value}",
            #       ))
            #   items.append(ci)
            # return items



"""A decorator represents the modification of the program around a function.
Consider the following example:

@decorator_name(decorator_arg, decorator_arg2)
def fun(...) { ... }

The decorator implementation will receive the list [decorator_arg, decorator_arg2] and the definition of fun.
Then the implementation should return a tuple of the (possibly modified) definition, as well as a list of
eventual complementary definitions.
"""

from aeon.decorators.api import DecoratorType
from aeon.decorators.api import Metadata
from aeon.sugar.program import Definition
from aeon.synthesis.decorators import (
    minimize_int,
    maximize_int,
    minimize_float,
    maximize_float,
    multi_minimize_float,
    hide,
    hide_types,
    allow_recursion,
    error_fitness,
    disable_control_flow,
    prompt,
)

decorators_environment: dict[str, DecoratorType] = {
    "minimize_int": minimize_int,
    "maximize_int": maximize_int,
    "minimize_float": minimize_float,
    "maximize_float": maximize_float,
    "multi_minimize_float": multi_minimize_float,
    "hide": hide,
    "hide_types": hide_types,
    "allow_recursion": allow_recursion,
    "error_fitness": error_fitness,
    "disable_control_flow": disable_control_flow,
    "prompt": prompt,
}


def apply_decorators(fun: Definition, metadata: Metadata) -> tuple[Definition, list[Definition], Metadata]:
    "Applies each decorator in order, and returns the cumulative list of possible new definitions."
    if not metadata:
        metadata = {}
    total_extra = []
    for decorator in fun.decorators:
        dname = decorator.name.name
        if dname not in decorators_environment:
            raise Exception(f"Unknown decorator named {dname}, in function {fun.name}.")
        decorator_processor = decorators_environment[dname]
        (fun, extra, metadata) = decorator_processor(decorator.macro_args, fun, metadata)
        total_extra.extend(extra)
    return fun, total_extra, metadata



from dataclasses import dataclass, field

from aeon.core.liquid import LiquidLiteralBool, LiquidTerm
from aeon.core.types import AbstractionType, TypeConstructor


class SMTSolvingException(Exception):
    pass


@dataclass
class SMTQuery:
    types: list[str] = field(default_factory=list)
    functions: dict[str, AbstractionType] = field(default_factory=lambda: {})
    variables: dict[str, TypeConstructor] = field(default_factory=lambda: {})
    premises: list[LiquidTerm] = field(default_factory=list)
    conclusion: LiquidTerm = field(default_factory=lambda: LiquidLiteralBool(True))



from __future__ import annotations

import datetime
import sys

from loguru import logger


def levels_filter(levels):
    def filter(record):
        return record["level"].name in levels

    return filter


def setup_logger():
    logger.level("TYPECHECKER", no=35, color="<magenta>", icon="🔍")
    logger.level("SYNTH_TYPE", no=36, color="<magenta>")
    logger.level("CONSTRAINT", no=37, color="<cyan>", icon="🔒")
    logger.level("SYNTHESIZER", no=38, color="<red>")
    logger.level("TIME", no=39, color="<green>")

    # Setup the logger
    logger.remove()
    # logger.add(sys.stderr, level="DEBUG")
    return logger


def export_log(logs: list, export_file: bool = False, logfile_name: str = ""):
    if export_file:
        logfile = f"logs/{logfile_name}_{datetime.datetime.now().strftime('%Y_%m_%d %H_%M_%S')}.log"
        return logger.add(logfile, filter=levels_filter(logs))
    else:
        return logger.add(sys.stderr, filter=levels_filter(logs))



from typing import MutableSequence
from aeon.elaboration.context import (
    ElabTypingContextEntry,
    ElabVariableBinder,
    ElaborationTypingContext,
    ElabUninterpretedBinder,
    ElabTypeVarBinder,
    ElabTypeDecl,
)
from aeon.sugar.program import (
    Decorator,
    Definition,
    Program,
    SAbstraction,
    SAnnotation,
    SApplication,
    SHole,
    SIf,
    SLet,
    SLiteral,
    SRec,
    STerm,
    STypeAbstraction,
    STypeApplication,
    SVar,
    TypeDecl,
)
from aeon.sugar.stypes import (
    SAbstractionType,
    SRefinedType,
    SType,
    STypeConstructor,
    STypePolymorphism,
    STypeVar,
)
from aeon.utils.name import Name, fresh_counter

RenamingSubstitions = list[tuple[str, Name]]


def get_last_id(x: str, subs: RenamingSubstitions) -> Name | None:
    for sub, v in subs[::-1]:
        if x == sub:
            return v
    return None


def check_name(name: Name, subs: RenamingSubstitions) -> tuple[Name, RenamingSubstitions]:
    if name.id == -1:
        nname = Name(name.name, fresh_counter.fresh())
        return nname, subs + [(name.name, nname)]
    else:
        return name, subs + [(name.name, name)]


def apply_subs_name(subs: RenamingSubstitions, name: Name) -> Name:
    for sub, v in subs[::-1]:
        if sub == name.name:
            return v
    return name


def bind_ectx(
    ectx: ElaborationTypingContext, subs: RenamingSubstitions
) -> tuple[ElaborationTypingContext, RenamingSubstitions]:
    nentries: MutableSequence[ElabTypingContextEntry] = []
    for entry in ectx.entries:
        e: ElabTypingContextEntry
        match entry:
            case ElabVariableBinder(name, stype):
                name, subs = check_name(name, subs)
                e = ElabVariableBinder(name, bind_stype(stype, subs))
            case ElabUninterpretedBinder(name, stype):
                name, subs = check_name(name, subs)
                e = ElabUninterpretedBinder(name, bind_stype(stype, subs))
            case ElabTypeVarBinder(name, kind):
                name, subs = check_name(name, subs)
                e = ElabTypeVarBinder(name, kind)
            case ElabTypeDecl(name, args):
                name, subs = check_name(name, subs)
                e = ElabTypeDecl(name, args)
            case _:
                assert False, f"{entry} not yet supported in bind."
        nentries.append(e)
    return ElaborationTypingContext(nentries), subs


def bind_stype(ty: SType, subs: RenamingSubstitions) -> SType:
    match ty:
        case STypeVar(name):
            return STypeVar(apply_subs_name(subs, name))
        case STypeConstructor(name, args):
            nargs = [bind_stype(aty, subs) for aty in args]
            return STypeConstructor(apply_subs_name(subs, name), nargs)
        case SAbstractionType(aname, atype, rtype):
            nname, nsubs = check_name(aname, subs)

            return SAbstractionType(nname, bind_stype(atype, subs), bind_stype(rtype, nsubs))
        case SRefinedType(name, ty, ref):
            nty = bind_stype(ty, subs)
            nname, nsubs = check_name(name, subs)
            nref = bind_sterm(ref, nsubs)
            return SRefinedType(nname, nty, nref)
        case STypePolymorphism(name, kind, body):
            name, subs = check_name(name, subs)
            return STypePolymorphism(name, kind, bind_stype(body, subs))
        case _:
            assert False, f"Unique not supported for {ty} ({type(ty)})"


def bind_sterm(t: STerm, subs: RenamingSubstitions) -> STerm:
    match t:
        case SLiteral(_, _):
            return t
        case SVar(name, loc=loc):
            return SVar(apply_subs_name(subs, name), loc=loc)
        case SHole(name, loc=loc):
            name, _ = check_name(name, subs)
            return SHole(name, loc=loc)
        case SAnnotation(e, ty, loc=loc):
            return SAnnotation(bind_sterm(e, subs), bind_stype(ty, subs), loc=loc)
        case SApplication(e1, e2, loc=loc):
            return SApplication(bind_sterm(e1, subs), bind_sterm(e2, subs), loc=loc)
        case SAbstraction(name, body, loc=loc):
            name, subs = check_name(name, subs)
            nbody = bind_sterm(body, subs)
            return SAbstraction(name, nbody, loc=loc)
        case STypeApplication(body, ty, loc=loc):
            return STypeApplication(bind_sterm(body, subs), bind_stype(ty, subs), loc=loc)
        case STypeAbstraction(name, kind, body, loc=loc):
            name, subs = check_name(name, subs)
            return STypeAbstraction(name, kind, bind_sterm(body, subs), loc=loc)
        case SIf(cond, then, otherwise, loc=loc):
            return SIf(bind_sterm(cond, subs), bind_sterm(then, subs), bind_sterm(otherwise, subs), loc=loc)
        case SLet(name, body, cont, loc=loc):
            name, nsubs = check_name(name, subs)
            return SLet(name, bind_sterm(body, subs), bind_sterm(cont, nsubs), loc=loc)
        case SRec(name, ty, body, cont, loc=loc):
            name, subs = check_name(name, subs)
            return SRec(name, bind_stype(ty, subs), bind_sterm(body, subs), bind_sterm(cont, subs), loc=loc)
        case _:
            assert False, f"Unique not supported for {t} ({type(t)})"


def bind_program(p: Program, subs: RenamingSubstitions) -> Program:
    type_decls = []
    definitions = []
    nsubs = list(subs)
    for td in p.type_decls:
        name, nsubs = check_name(td.name, nsubs)
        type_decls.append(TypeDecl(name, td.args, loc=td.loc))
    for df in p.definitions:
        name, nsubs = check_name(df.name, nsubs)
        foralls = []
        for name, kind in df.foralls:
            nname, nsubs = check_name(name, nsubs)
            foralls.append((nname, kind))
        args = []
        for aname, ty in df.args:
            nname, nsubs = check_name(aname, nsubs)
            ty = bind_stype(ty, nsubs)
            args.append((nname, ty))
        ntype = bind_stype(df.type, nsubs)
        body = bind_sterm(df.body, nsubs)
        decorators = []
        for dec in df.decorators:
            dargs = []
            for da in dec.macro_args:
                dargs.append(bind_sterm(da, subs))
            decorators.append(Decorator(dec.name, dargs))
        d = Definition(name, foralls, args, ntype, body, decorators, loc=df.loc)
        definitions.append(d)
    return Program(p.imports, type_decls, [], definitions)


def bind(ectx: ElaborationTypingContext, s: STerm) -> tuple[ElaborationTypingContext, STerm]:
    subs: RenamingSubstitions = []
    ectx, subs = bind_ectx(ectx, subs)
    return ectx, bind_sterm(s, subs)



from abc import ABC, abstractmethod
from dataclasses import dataclass


class Location(ABC):
    @abstractmethod
    def get_start(self) -> tuple[int, int]:
        pass

    @abstractmethod
    def get_end(self) -> tuple[int, int]:
        pass


@dataclass(unsafe_hash=True, frozen=True)
class FileLocation(Location):
    file: str
    start: tuple[int, int]
    end: tuple[int, int]

    def get_start(self) -> tuple[int, int]:
        return self.start

    def get_end(self) -> tuple[int, int]:
        return self.end


@dataclass(unsafe_hash=True, frozen=True)
class SynthesizedLocation(Location):
    reason: str

    def get_start(self) -> tuple[int, int]:
        return 0, 0

    def get_end(self) -> tuple[int, int]:
        return 0, 0



from itertools import count, takewhile
import itertools
from typing import Callable

from aeon.core.terms import Annotation, Application, If, Literal, TypeApplication, Var
from aeon.core.types import AbstractionType, RefinedType, Type, TypeConstructor, TypePolymorphism, TypeVar
from aeon.typechecking.context import TypingContext
from aeon.utils.name import Name


def monomorfic(t: Type, typing_ctx: TypingContext, t_l: dict[Name, Type]):
    match t:
        case TypePolymorphism(var_name, _, var_body):
            for t in [
                t
                for n, t in typing_ctx.concrete_vars()
                if isinstance(t, TypeConstructor) and t != TypeConstructor(Name("Unit", 0))
            ]:
                t_l[var_name] = t
                for result in monomorfic(var_body, typing_ctx, t_l):
                    yield result
        case AbstractionType(var_name, var_type, typ):
            for in_type in monomorfic(var_type, typing_ctx, t_l):
                for body in monomorfic(typ, typing_ctx, t_l):
                    yield AbstractionType(var_name, in_type, body)
        case TypeVar(var_name):
            yield t_l[var_name]
        case _:
            yield t


def uncurry(typ: Type):
    match typ:
        case TypeConstructor():
            return [], typ
        case TypeVar(var_name):
            return [], TypeConstructor(var_name, [])
        case AbstractionType(_, _, _):
            t = typ
            params = []
            while isinstance(t, AbstractionType):
                if isinstance(t.var_type, (TypeConstructor, AbstractionType)):
                    params.append(t.var_type)
                else:
                    assert isinstance(t.var_type, RefinedType)
                    if isinstance(t.var_type.type, TypeConstructor):
                        params.append(t.var_type.type)
                if not isinstance(t.type, AbstractionType):
                    break
                t = t.type

            if isinstance(t.type, TypeConstructor):
                return params, t.type
            else:
                assert isinstance(t.type, RefinedType)
                return params, t.type.type
        case RefinedType():
            return [], typ.type
        case _:
            print(type(typ))
            assert False, f"Unsupported {typ}"


def frange(start, stop, step):
    return takewhile(lambda x: x < stop, count(start, step))


def closing(elems: tuple, typ: TypeConstructor):
    if len(elems) == 1:
        return TypeApplication(elems[0], typ)
    else:
        return Application(closing(elems[:-1], typ), elems[-1])


def synthes_memory(ctx: TypingContext, level: int, ret_t: Type, skip: Callable[[Name], bool], mem: dict):
    if (ctx, level, ret_t) in mem:
        yield from mem[(ctx, level, ret_t)]
    else:
        mem[(ctx, level, ret_t)] = []
        try:
            for item in synthes(ctx, level, ret_t, skip, mem):
                mem[(ctx, level, ret_t)].append(item)
                yield item
        except NotImplementedError:
            yield from []


def match_type(t1: Type, t2: Type):
    if t1 == t2:
        return True
    elif type(t1) is type(t2):
        return False
    elif isinstance(t1, AbstractionType) and isinstance(t2, AbstractionType):
        return match_type(t1.var_type, t2.var_type) and match_type(t1.type, t2.type)
    else:
        return False


def synthes(ctx: TypingContext, level: int, ret_t: Type, skip: Callable[[Name], bool], mem: dict):
    if level == 0:
        if isinstance(ret_t, AbstractionType):
            for name, _ in [
                (n, t) for n, t in ctx.concrete_vars() if isinstance(t, AbstractionType) and match_type(t, ret_t)
            ]:
                yield Var(name)
        else:
            assert isinstance(ret_t, TypeConstructor)
            for name, _ in [
                (n, t) for n, t in ctx.concrete_vars() if isinstance(t, TypeConstructor) and match_type(t, ret_t)
            ]:
                yield Var(name)
        match ret_t:
            case TypeConstructor(Name("Bool", 0)):
                yield from [Literal(True, ret_t), Literal(False, ret_t)]
            case TypeConstructor(Name("Int", 0)):
                yield from [Literal(value, ret_t) for value in range(-100, 100)]
            case TypeConstructor(Name("Float", 0)):
                yield from [Literal(value, ret_t) for value in frange(-100.0, 100.0, 0.00001)]
            case TypeConstructor(Name("String", 0)):
                raise NotImplementedError
            case _:
                raise NotImplementedError
    elif level >= 1:
        for name, typ in [
            (n, ty) for n, ty in ctx.concrete_vars() if not isinstance(ty, TypeConstructor) and not skip(n)
        ]:
            for candidate in monomorfic(typ, ctx, {}) if isinstance(typ, TypePolymorphism) else [typ]:
                params_t, t = uncurry(candidate)
                if t != ret_t:
                    continue
                params = [
                    synthes_memory(ctx, level - 1, i, skip, mem)
                    if isinstance(i, TypeConstructor)
                    else synthes_memory(ctx, 0, i, skip, mem)
                    for i in params_t
                ]
                params.insert(0, [Var(name)])
                for i in itertools.product(*params):
                    a = closing(i, t)
                    yield a
        # if
        cond = synthes_memory(ctx, level - 1, TypeConstructor(Name("Bool", 0)), skip, mem)
        then = synthes_memory(ctx, level - 1, ret_t, skip, mem)
        otherwise = synthes_memory(ctx, level - 1, ret_t, skip, mem)
        for cand in itertools.product(*[cond, then, otherwise]):
            if isinstance(cand[0], If):
                break
            yield Annotation(If(cand[0], cand[1], cand[2]), ret_t)



from typing import Any, Callable

from aeon.typechecking.context import TypingContext

from aeon.core.types import Type
from aeon.core.terms import Term
from aeon.utils.name import Name


from aeon.synthesis.uis.api import SynthesisUI
from aeon.synthesis.api import Synthesizer

from aeon.decorators.api import Metadata

from aeon.synthesis.grammar.grammar_generation import create_grammar

from geneticengine.evaluation.budget import TimeBudget
from geneticengine.problems import InvalidFitnessException
from geneticengine.problems import MultiObjectiveProblem, Problem, SingleObjectiveProblem
from geneticengine.representations.tree.treebased import TreeBasedRepresentation
from geneticengine.random.sources import NativeRandomSource
from geneticengine.representations.tree.initializations import MaxDepthDecider
from geneticengine.evaluation import SequentialEvaluator
from geneticengine.evaluation.tracker import ProgressTracker
from geneticengine.algorithms.random_search import RandomSearch
from geneticengine.algorithms.enumerative import EnumerativeSearch
from geneticengine.algorithms.gp.gp import GeneticProgramming
from geneticengine.algorithms.one_plus_one import OnePlusOne
from geneticengine.algorithms.hill_climbing import HC
from geneticengine.solutions import Individual
from geneticengine.evaluation.recorder import SearchRecorder

from aeon.synthesis.decorators import Goal


def create_problem(
    validate: Callable[[Term], bool],
    evaluate: Callable[[Term], list[float]],
    fun_name: Name,
    metadata: Metadata,
) -> Problem:
    goals: list[Goal] = metadata.get(fun_name, {}).get("goals", [])
    minimize_list = [goal.minimize for goal in goals for _ in range(goal.length)]
    if minimize_list:

        def fitness_fun(phenotype: Any) -> list[float]:
            p = phenotype.get_core()
            assert isinstance(p, Term)
            if not validate(p):
                raise InvalidFitnessException()
            return evaluate(p)

        return MultiObjectiveProblem(fitness_function=fitness_fun, minimize=minimize_list)
    else:

        def single_fitness_fun(phenotype: Any) -> float:
            p = phenotype.get_core()
            assert isinstance(p, Term)
            if not validate(p):
                raise InvalidFitnessException()
            return 0.0

        return SingleObjectiveProblem(fitness_function=single_fitness_fun, minimize=False, target=0.0)


class GESynthesizer(Synthesizer):
    def __init__(self, seed: int = 0, method: str = "enumerative"):
        self.seed = seed
        self.method = method

    def synthesize(
        self,
        ctx: TypingContext,
        type: Type,
        validate: Callable[[Term], bool],
        evaluate: Callable[[Term], list[float]],
        fun_name: Name,
        metadata: Metadata,
        budget: float = 60,
        ui: SynthesisUI = SynthesisUI(),
    ) -> Term:
        assert isinstance(ctx, TypingContext)
        assert isinstance(type, Type)

        class UIBackendRecorder(SearchRecorder):
            def register(self, tracker: Any, individual: Individual, problem: Problem, is_best=False):
                ui.register(
                    individual.get_phenotype().get_core(),
                    individual.get_fitness(problem),
                    tracker.get_elapsed_time(),
                    is_best,
                )

        grammar = create_grammar(ctx, type, fun_name, metadata)
        problem = create_problem(validate, evaluate, fun_name, metadata)

        tracker = ProgressTracker(problem, evaluator=SequentialEvaluator(), recorders=[UIBackendRecorder()])

        representation = TreeBasedRepresentation(
            grammar, decider=MaxDepthDecider(NativeRandomSource(self.seed), grammar, max_depth=5)
        )

        common_args = {
            "problem": problem,
            "budget": TimeBudget(budget),
            "tracker": tracker,
        }

        common_random_args = {"representation": representation, "random": NativeRandomSource(self.seed), **common_args}

        match self.method:
            case "random_search":
                alg = RandomSearch(**common_random_args)
            case "enumerative":
                alg = EnumerativeSearch(grammar=grammar, **common_args)
            case "genetic_programming":
                alg = GeneticProgramming(**common_random_args)
            case "hill_climbing":
                alg = HC(**common_random_args)
            case "one_plus_one":
                alg = OnePlusOne(**common_random_args)
            case _:
                assert False, f"Method {self.method} not available for synthesis."
        individuals = alg.search()
        match individuals:
            case None | [None]:
                return None
            case [ind, *_]:
                # TODO: handle multiple answers
                if not ind.get_fitness(problem).valid:
                    return None
                return ind.get_phenotype().get_core()
            case _:
                return None



import abc
import json
from enum import Enum
from typing import Any

from aeon.backend.evaluator import EvaluationContext
from aeon.core.terms import Term
from aeon.core.types import Type
from aeon.sugar.program import STerm
from aeon.typechecking.context import TypingContext
from aeon.utils.name import Name
from aeon.utils.pprint import pretty_print_sterm


class SynthesisFormat(Enum):
    DEFAULT = "default"
    JSON = "json"

    @classmethod
    def from_string(cls, string_value):
        for member in cls:
            if member.value == string_value:
                return member
        return cls.DEFAULT


class SynthesisUI(abc.ABC):
    def start(
        self,
        typing_ctx: TypingContext,
        evaluation_ctx: EvaluationContext,
        target_name: str,
        target_type: Type | None,
        budget: Any,
    ): ...

    def register(
        self,
        solution: Term,
        quality: Any,
        elapsed_time: float,
        is_best: bool,
    ): ...

    def end(self, solution: Term, quality: Any): ...

    def display_results(
        self,
        program: Term,
        terms: dict[Name, STerm],
        synthesis_format: SynthesisFormat = SynthesisFormat.DEFAULT,
    ):
        print("Synthesized holes:")
        match synthesis_format:
            case SynthesisFormat.JSON:
                result = {
                    f"?{name.pretty()}": pretty_print_sterm(terms[name]) if name in terms else "None" for name in terms
                }
                print(json.dumps(result, indent=2))

            case _:
                for name in terms:
                    name_str = name.pretty()
                    term_str = pretty_print_sterm(terms[name]) if name in terms else "None"
                    print(f"?{name_str}: {term_str}")
        # print()
        # pretty_print_term(ensure_anf(synthesis_result, 200))


class SilentSynthesisUI(SynthesisUI):
    def start(
        self,
        typing_ctx: TypingContext,
        evaluation_ctx: EvaluationContext,
        target_name: str,
        target_type: Type,
        budget: Any,
    ):
        pass

    def register(
        self,
        solution: Term,
        quality: Any,
        elapsed_time: float,
        is_best: bool,
    ):
        pass

    def end(self, solution: Term, quality: Any):
        pass



# TODO: Do this for all AST
from dataclasses import dataclass

from lark import Token


@dataclass
class SourcePosition:
    line_start: int
    line_end: int
    col_start: int
    col_ind: int

    @staticmethod
    def fromToken(t: Token):
        return SourcePosition(t.line, t.end_line, t.column, t.end_column)



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass, field
from aeon.utils.location import Location, SynthesizedLocation
from aeon.utils.name import Name


class LiquidTerm(ABC):
    loc: Location


def ensure_liqterm(a: LiquidTerm | Name) -> LiquidTerm:
    if isinstance(a, Name):
        return LiquidVar(a)
    return a


class LiquidHole(LiquidTerm):
    def __eq__(self, other):
        return isinstance(other, self.__class__)


def is_safe_for_application(x: LiquidTerm):
    return (
        isinstance(x, LiquidVar)
        or isinstance(x, LiquidLiteralBool)
        or isinstance(x, LiquidLiteralFloat)
        or isinstance(x, LiquidLiteralInt)
        or isinstance(x, LiquidLiteralString)
    )


@dataclass
class LiquidLiteralBool(LiquidTerm):
    value: bool
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __repr__(self):
        return f"{self.value}".lower()

    def __eq__(self, other):
        return isinstance(other, LiquidLiteralBool) and other.value == self.value

    def __hash__(self) -> int:
        return hash(self.value)


@dataclass
class LiquidLiteralInt(LiquidTerm):
    value: int
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __repr__(self):
        return f"{self.value}".lower()

    def __eq__(self, other):
        return isinstance(other, LiquidLiteralInt) and other.value == self.value

    def __hash__(self) -> int:
        return hash(self.value)


@dataclass
class LiquidLiteralFloat(LiquidTerm):
    value: float
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __repr__(self):
        return f"{self.value}".lower()

    def __eq__(self, other):
        return isinstance(other, LiquidLiteralFloat) and other.value == self.value

    def __hash__(self) -> int:
        return hash(self.value)


@dataclass
class LiquidLiteralString(LiquidTerm):
    value: str
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __repr__(self):
        return f"{self.value}".lower()

    def __eq__(self, other):
        return isinstance(other, LiquidLiteralString) and other.value == self.value

    def __hash__(self) -> int:
        return hash(self.value)


@dataclass
class LiquidVar(LiquidTerm):
    name: Name
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __repr__(self):
        return f"{self.name}"

    def __eq__(self, other):
        return isinstance(other, LiquidVar) and other.name == self.name

    def __hash__(self) -> int:
        return hash(self.name)


@dataclass
class LiquidApp(LiquidTerm):
    fun: Name
    args: list[LiquidTerm]
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __repr__(self):
        if all([not c.isalnum() for c in self.fun.name]) and len(self.args) == 2:
            (a1, a2) = (repr(x) for x in self.args)
            return f"({a1} {self.fun} {a2})"

        fargs = ",".join([repr(x) for x in self.args])
        return f"{self.fun}({fargs})"

    def __eq__(self, other):
        return (
            isinstance(other, LiquidApp)
            and other.fun == self.fun
            and all(x == y for (x, y) in zip(self.args, other.args))
        )

    def __hash__(self) -> int:
        return hash(self.fun) + sum(hash(a) for a in self.args)


def liquid_free_vars(e: LiquidTerm) -> list[Name]:
    if isinstance(e, LiquidVar):
        return [e.name]
    elif isinstance(e, LiquidApp):
        return [e.fun] + [x for arg in e.args for x in liquid_free_vars(arg)]
    else:
        return []



from __future__ import annotations

import importlib
from typing import Any

from aeon.sugar.parser import parse_type
from aeon.sugar.stypes import SType
from aeon.utils.name import Name

INTEGER_ARITHMETIC_OPERATORS = ["+", "*", "-", "/", "%"]
COMPARISON_OPERATORS = ["<", ">", "<=", ">="]
LOGICAL_OPERATORS = ["&&", "||"]
EQUALITY_OPERATORS = ["==", "!="]

ALL_OPS = INTEGER_ARITHMETIC_OPERATORS + COMPARISON_OPERATORS + LOGICAL_OPERATORS + EQUALITY_OPERATORS


def p(x):
    print(str(x))
    return 0


def native_import(name):
    return importlib.import_module(name)


native_types: list[Name] = [Name("Unit", 0), Name("Bool", 0), Name("Int", 0), Name("Float", 0), Name("String", 0)]

# TODO: polymorphic signatures
prelude = [
    ("native", "forall a:B, (x:String) -> {x:a | false}", eval),
    ("native_import", "forall a:B, (x:String) -> {x:a | false}", native_import),
    ("print", "forall a:B, (x:a) -> Unit", p),
    ("==", "forall a:B, (x:a) -> (y:a) -> Bool", lambda x: lambda y: x == y),
    ("!=", "forall a:B, (x:a) -> (y:a) -> Bool", lambda x: lambda y: x != y),
    ("<", "forall a:B, (x:a) -> (y:a) -> Bool", lambda x: lambda y: x < y),
    ("<=", "forall a:B, (x:a) -> (y:a) -> Bool", lambda x: lambda y: x <= y),
    (">", "forall a:B, (x:a) -> (y:a) -> Bool", lambda x: lambda y: x > y),
    (">=", "forall a:B, (x:a) -> (y:a) -> Bool", lambda x: lambda y: x >= y),
    ("+", "forall a:B, (x:a) -> (y:a) -> a", lambda x: lambda y: x + y),
    ("-", "forall a:B, (x:a) -> (y:a) -> a", lambda x: lambda y: x - y),
    ("*", "forall a:B, (x:a) -> (y:a) -> a", lambda x: lambda y: x * y),
    ("/", "forall a:B, (x:a) -> (y:a) -> a", lambda x: lambda y: x / y),
    ("%", "(x:Int) -> (y:Int) -> Int", lambda x: lambda y: x % y),
    ("&&", "(x:Bool) -> (y:Bool) -> Bool", lambda x: lambda y: x and y),
    ("||", "(x:Bool) -> (y:Bool) -> Bool", lambda x: lambda y: x or y),
    ("!", "(x:Bool) -> Bool", lambda x: not x),
]

typing_vars: dict[Name, SType] = {}
evaluation_vars: dict[Name, Any] = {}


for n, ty, ex in prelude:
    nn = Name(n, 0)
    typing_vars[nn] = parse_type(ty)
    evaluation_vars[nn] = ex



from __future__ import annotations

import sys
from typing import Annotated, Any
from typing import Union

from geneticengine.grammar.metahandlers.base import MetaHandlerGenerator
from geneticengine.grammar.metahandlers.vars import VarRange


from sympy.core.numbers import Infinity, NegativeInfinity
from sympy.sets.sets import Interval, EmptySet, FiniteSet
from sympy.simplify.simplify import simplify
from sympy.logic.boolalg import to_dnf

from aeon.core.types import Type
from aeon.core.types import RefinedType
from aeon.synthesis.grammar.bounds import (
    conditional_to_interval,
    flatten_conditions,
    refined_to_sympy_expression,
    sympy_exp_to_bounded_interval,
)
from aeon.synthesis.grammar.utils import (
    aeon_to_gengy_metahandlers,
    aeon_to_python,
)

max_number = sys.maxsize - 1
min_number = -(sys.maxsize - 1)


def replace_tuples_with_lists(structure):
    if isinstance(structure, tuple):
        return [replace_tuples_with_lists(item) for item in structure]
    elif isinstance(structure, list):
        return [replace_tuples_with_lists(item) for item in structure]
    else:
        return structure


def contains_tuples(lst):
    return any(isinstance(item, tuple) for item in lst)


def split_or_intervals(bounded_intervals, name: str, intervals_list=None):
    intervals_list = [] if intervals_list is None else intervals_list
    # if it is a tuple, it is an Or Interval
    if isinstance(bounded_intervals, tuple):
        for b_interval in bounded_intervals:
            intervals_list = split_or_intervals(b_interval, name, intervals_list)
    elif isinstance(bounded_intervals, list):
        if contains_tuples(bounded_intervals):
            for b_interval in bounded_intervals:
                intervals_list = split_or_intervals(b_interval, name, intervals_list)
        else:
            cond = flatten_conditions(bounded_intervals)
            interval = conditional_to_interval(cond, name)
            intervals_list.append(interval)
    return intervals_list


def interval_to_metahandler(interval: Any, base_type: Type) -> MetaHandlerGenerator:
    python_type: type = aeon_to_python[base_type]
    match interval:
        case EmptySet():
            return None
        case FiniteSet():
            values = [python_type(x) for x in interval.args]
            inst = VarRange(values)
            return Annotated[python_type, inst]
        case Interval():
            max_range = (
                max_number if isinstance(interval.sup, Infinity) else python_type(interval.sup)
            )  # or 2 ** 31 - 1
            max_range = max_range - 1 if interval.right_open else max_range

            min_range = (
                min_number if isinstance(interval.inf, NegativeInfinity) else python_type(interval.inf)
            )  # or -2 ** 31
            min_range = min_range + 1 if interval.left_open else min_range

            metahandler_instance = aeon_to_gengy_metahandlers[base_type](min_range, max_range)
            return Annotated[python_type, metahandler_instance]
        case _:
            raise NotImplementedError()


def intervals_to_metahandlers(intervals_list: list, base_type: Type) -> list[MetaHandlerGenerator]:
    return [x for x in [interval_to_metahandler(i, base_type) for i in intervals_list] if x is not None]


def get_metahandler_union(
    metahandler_list: list[MetaHandlerGenerator],
) -> MetaHandlerGenerator | Union[MetaHandlerGenerator]:
    if len(metahandler_list) == 1:
        return metahandler_list[0]
    else:
        return Union[tuple(metahandler_list)]


def refined_type_to_metahandler(ty: RefinedType) -> MetaHandlerGenerator | Union[MetaHandlerGenerator]:
    name, ref = str(ty.name), ty.refinement
    sympy_exp = refined_to_sympy_expression(ref)
    sympy_exp = simplify(sympy_exp)
    sympy_exp = to_dnf(sympy_exp)
    bounded_intervals = sympy_exp_to_bounded_interval(sympy_exp)
    intervals_list = split_or_intervals(bounded_intervals, name)
    metahandler_list = intervals_to_metahandlers(intervals_list, ty.type)

    return get_metahandler_union(metahandler_list)



from __future__ import annotations

from dataclasses import dataclass, field

from aeon.core.types import Kind
from aeon.core.types import t_string
from aeon.core.types import Type
from aeon.utils.location import Location, SynthesizedLocation
from aeon.utils.name import Name


class Term:
    loc: Location

    def __hash__(self) -> int:
        return str(self).__hash__()

    def pretty(self):
        pass


@dataclass(frozen=True)
class Literal(Term):
    value: object
    type: Type
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __repr__(self):
        if self.type == t_string:
            return f'"{self.value}"'
        return f"{self.value}".lower()

    def __eq__(self, other):
        return isinstance(other, Literal) and self.value == other.value and self.type == other.type

    def pretty(self):
        if self.type == t_string:
            return f'"{self.value}"'
        return f"{self.value}".lower()


@dataclass(frozen=True)
class Var(Term):
    name: Name
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"{self.name}"

    def __repr__(self):
        return f"Var({self.name})"

    def __eq__(self, other):
        return isinstance(other, Var) and self.name == other.name

    def pretty(self):
        return self.name.pretty()


@dataclass(frozen=True)
class Annotation(Term):
    expr: Term
    type: Type
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"({self.expr} : {self.type})"

    def __repr__(self):
        return f"({self.expr} : {self.type})"

    def __eq__(self, other):
        return isinstance(other, Annotation) and self.expr == other.expr


@dataclass(frozen=True)
class Hole(Term):
    name: Name
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"?{self.name}"

    def __repr__(self):
        return f"Hole({self.name})"

    def __eq__(self, other):
        return isinstance(other, Hole) and self.name == other.name


@dataclass(frozen=True)
class Application(Term):
    fun: Term
    arg: Term
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __repr__(self):
        return f"({self.fun} {self.arg})"

    def __eq__(self, other):
        return isinstance(other, Application) and self.fun == other.fun and self.arg == other.arg


@dataclass(frozen=True)
class Abstraction(Term):
    var_name: Name
    body: Term
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __post_init__(self):
        assert isinstance(self.var_name, Name)

    def __repr__(self):
        return f"(\\{self.var_name} -> {self.body})"

    def __eq__(self, other):
        return isinstance(other, Abstraction) and self.var_name == other.var_name and self.body == other.body


@dataclass(frozen=True)
class Let(Term):
    var_name: Name
    var_value: Term
    body: Term
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"(let {self.var_name} = {self.var_value} in\n\t{self.body})"

    def __eq__(self, other):
        return (
            isinstance(other, Let)
            and self.var_name == other.var_name
            and self.var_value == other.var_value
            and self.body == other.body
        )


@dataclass(frozen=True)
class Rec(Term):
    var_name: Name
    var_type: Type
    var_value: Term
    body: Term
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "(let {} : {} = {} in\n\t{})".format(
            self.var_name,
            self.var_type,
            self.var_value,
            self.body,
        )

    def __eq__(self, other):
        return (
            isinstance(other, Rec)
            and self.var_name == other.var_name
            and self.var_type == other.var_type
            and self.var_value == other.var_value
            and self.body == other.body
        )


@dataclass(frozen=True)
class If(Term):
    cond: Term
    then: Term
    otherwise: Term
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"(if {self.cond} then {self.then} else {self.otherwise})"

    def __eq__(self, other):
        return (
            isinstance(other, If)
            and self.cond == other.cond
            and self.then == other.then
            and self.otherwise == other.otherwise
        )


@dataclass(frozen=True)
class TypeAbstraction(Term):
    name: Name
    kind: Kind
    body: Term
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"ƛ{self.name}:{self.kind}.({self.body})"


@dataclass(frozen=True)
class TypeApplication(Term):
    body: Term
    type: Type
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"({self.body})[{self.type}]"



from __future__ import annotations

import os
from pathlib import Path
from typing import NamedTuple

from aeon.core.types import BaseKind, Kind
from aeon.decorators import apply_decorators, Metadata
from aeon.elaboration.context import (
    ElabUninterpretedBinder,
    ElabVariableBinder,
    ElaborationTypingContext,
    ElabTypingContextEntry,
    build_typing_context,
)
from aeon.prelude.prelude import typing_vars
from aeon.sugar.parser import mk_parser
from aeon.sugar.program import (
    Definition,
    SAbstraction,
    SApplication,
    SHole,
    SLiteral,
    SRec,
    STerm,
    STypeAbstraction,
    SVar,
)
from aeon.sugar.program import ImportAe
from aeon.sugar.program import Program
from aeon.sugar.program import TypeDecl, InductiveDecl
from aeon.sugar.stypes import SAbstractionType, SType, STypePolymorphism, builtin_types, get_type_vars, STypeConstructor
from aeon.sugar.substitutions import substitute_svartype_in_stype, substitution_svartype_in_sterm
from aeon.utils.name import Name
from aeon.sugar.ast_helpers import st_int, st_string

from aeon.sugar.stypes import STypeVar
from aeon.facade.api import ImportError


class DesugaredProgram(NamedTuple):
    program: STerm
    elabcontext: ElaborationTypingContext
    metadata: Metadata


def desugar(p: Program, is_main_hole: bool = True, extra_vars: dict[Name, SType] | None = None) -> DesugaredProgram:
    vs: dict[Name, SType] = {} if extra_vars is None else extra_vars
    vs.update(typing_vars)

    p = expand_inductive_decls(p)

    prog = determine_main_function(p, is_main_hole)

    defs, type_decls = p.definitions, p.type_decls
    defs, type_decls = handle_imports(p.imports, defs, type_decls)

    defs, metadata = apply_decorators_in_definitions(defs)

    defs = introduce_forall_in_types(defs, type_decls)

    etctx = build_typing_context(vs, type_decls)
    etctx, prog = update_program_and_context(prog, defs, etctx)
    prog, etctx = replace_concrete_types(
        prog, etctx, [Name(t, 0) for t in builtin_types] + [td.name for td in type_decls]
    )
    return DesugaredProgram(prog, etctx, metadata)


def expand_inductive_decls(p: Program) -> Program:
    tds: list[TypeDecl] = []
    defs: list[Definition] = []

    uninterpreted_lit = SVar(Name("uninterpreted", 0))

    for decl in p.inductive_decls:
        match decl:
            case InductiveDecl(name, args, constructors, measures, loc):
                tds.append(TypeDecl(name, args))

                for measure in measures:
                    match measure:
                        case Definition(mname, mforalls, margs, mrtype, _, _):
                            de = Definition(mname, mforalls, margs, mrtype, uninterpreted_lit)
                            defs.append(de)

                def key_for(tyname: Name, constructor_name: Name) -> str:
                    return f"{tyname.name}_{constructor_name.name}"

                for constructor in constructors:
                    match constructor:
                        case Definition(cname, cforalls, cargs, crtype, _, _):
                            arg_s = ", ".join(str(arg.name) for (arg, _) in cargs)
                            mk_tuple = SApplication(
                                SVar(Name("native", 0)), SLiteral(f"('{key_for(name, cname)}', {arg_s})", st_string)
                            )
                            de = Definition(cname, cforalls, cargs, crtype, mk_tuple)
                            defs.append(de)

                def curry(args: list[tuple[Name, SType]], rty: SType) -> SType:
                    for aname, aty in args[::-1]:
                        rty = SAbstractionType(aname, aty, rty)
                    return rty

                def case_for(cname: Name, cargs: list[tuple[Name, SType]]) -> str:
                    pargs = ", ".join(f"{arg.name}" for (arg, _) in cargs)
                    args = "".join(f"({arg.name})" for (arg, _) in cargs)
                    return f"\tcase ('{key_for(name, cname)}', {pargs}):\n\t\treturn case_{cname.name}{args}"

                cases = "\n".join(case_for(cons.name, cons.args) for cons in constructors)
                catchall = "\n\tcase _:\n\t\traise Exception('Invalid constructor')"
                rec_body: STerm = SApplication(
                    SVar(Name("native", 0)), SLiteral(f"""match this:\n{cases}{catchall}\n""", st_string)
                )

                foralls: list[tuple[Name, Kind]] = []
                rec_args: list[tuple[Name, SType]] = []

                # Return Type
                return_generic_name = Name("ret", -1)
                return_type = STypeVar(return_generic_name)
                foralls.append((return_generic_name, BaseKind()))

                # Target Type (First argument)
                # foralls.extend([ (arg, BaseKind()) for arg in args ])
                target_type = STypeConstructor(name, [STypeVar(a) for a in args])
                rec_args.append((Name("this", -1), target_type))

                # Prepare arguments for each constructor
                for cons in constructors:
                    rec_args.append((Name(f"case_{cons.name.name}", -1), curry(cons.args, return_type)))

                rec_de = Definition(
                    name=Name(name.name + "_rec", -1),
                    foralls=foralls,
                    args=rec_args,
                    type=return_type,
                    body=rec_body,
                    loc=loc,
                )
                defs.append(rec_de)

            case _:
                assert False, f"Unexpected inductive decl {decl} in {p}"

    return Program(p.imports, p.type_decls + tds, [], defs + p.definitions)


def introduce_forall_in_types(defs: list[Definition], type_decls: list[TypeDecl]) -> list[Definition]:
    types = [td.name for td in type_decls]
    ndefs = []
    for d in defs:
        match d:
            case Definition(name, foralls, args, rtype, body, decorators, loc):
                new_foralls: list[tuple[Name, Kind]] = []

                tlst: list[SType] = [ty for _, ty in args] + [rtype]
                for ty in tlst:
                    for t in get_type_vars(ty):
                        tname = t.name
                        if tname not in types:
                            entry = (tname, BaseKind())
                            if entry not in new_foralls:
                                new_foralls.append(entry)
                ndefs.append(Definition(name, foralls + new_foralls, args, rtype, body, decorators, loc))
    return ndefs


def determine_main_function(p: Program, is_main_hole: bool = True) -> STerm:
    for d in p.definitions:
        match d.name:
            case Name("main", id):
                return SApplication(SVar(Name("main", id)), SLiteral(1, type=st_int))
    if is_main_hole:
        return SHole(Name("main", 0))
    else:
        return SLiteral(1, st_int)


def handle_imports(
    imports: list[ImportAe],
    defs: list[Definition],
    type_decls: list[TypeDecl],
) -> tuple[list[Definition], list[TypeDecl]]:
    for imp in imports[::-1]:
        import_p = handle_import(imp)
        import_p_definitions = import_p.definitions
        defs_recursive: list[Definition] = []
        type_decls_recursive: list[TypeDecl] = []
        if import_p.imports:
            defs_recursive, type_decls_recursive = handle_imports(
                import_p.imports,
                import_p.definitions,
                import_p.type_decls,
            )
        if imp.func:
            import_p_definitions = [d for d in import_p_definitions if str(d.name.name) == imp.func]

        defs = defs_recursive + import_p_definitions + defs
        type_decls = type_decls_recursive + import_p.type_decls + type_decls
    return defs, type_decls


def apply_decorators_in_program(prog: Program) -> Program:
    """We apply the decorators meta-programming code to each definition in the
    program."""
    defs, _ = apply_decorators_in_definitions(prog.definitions)
    return Program(
        imports=prog.imports,
        type_decls=prog.type_decls,
        inductive_decls=prog.inductive_decls,
        definitions=defs,
    )


def apply_decorators_in_definitions(definitions: list[Definition]) -> tuple[list[Definition], Metadata]:
    """We apply the decorators meta-programming code to each definition in the
    program."""
    metadata: Metadata = {}
    new_definitions = []
    for definition in definitions:
        new_def, other_defs, metadata = apply_decorators(definition, metadata)
        new_definitions.append(new_def)
        new_definitions.extend(other_defs)
    return new_definitions, metadata


def update_program_and_context(
    prog: STerm,
    defs: list[Definition],
    ctx: ElaborationTypingContext,
) -> tuple[ElaborationTypingContext, STerm]:
    for d in defs[::-1]:
        match d.body:
            case SVar(Name("uninterpreted", _)):
                ctx.entries.append(ElabUninterpretedBinder(d.name, type_of_definition(d)))
            case _:
                prog = convert_definition_to_srec(prog, d)
    return ctx, prog


def replace_concrete_types(
    t: STerm, etctx: ElaborationTypingContext, types: list[Name]
) -> tuple[STerm, ElaborationTypingContext]:
    """Replaces all occurrences of STypeVar with the corresponding STypeConstructor."""
    for name in types:
        t = substitution_svartype_in_sterm(t, STypeConstructor(name), name)

    def fix_vartype(e: ElabTypingContextEntry) -> ElabTypingContextEntry:
        match e:
            case ElabVariableBinder(vname, ty):
                nty = ty
                for name in types:
                    nty = substitute_svartype_in_stype(nty, STypeConstructor(name), name)
                return ElabVariableBinder(vname, nty)
            case ElabUninterpretedBinder(vname, ty):
                nty = ty
                for name in types:
                    nty = substitute_svartype_in_stype(nty, STypeConstructor(name), name)
                return ElabUninterpretedBinder(vname, nty)
            case _:
                return e

    return t, ElaborationTypingContext([fix_vartype(e) for e in etctx.entries])


def type_of_definition(d: Definition) -> SType:
    match d:
        case Definition(_, foralls, args, rtype, _, _, loc):
            ntype = rtype
            for name, atype in reversed(args):
                ntype = SAbstractionType(name, atype, ntype, loc)
            for name, kind in reversed(foralls):
                ntype = STypePolymorphism(name, kind, ntype, loc)
            return ntype
        case _:
            assert False, f"{d} is not a definition"


def convert_definition_to_srec(prog: STerm, d: Definition) -> STerm:
    match d:
        case Definition(dname, foralls, args, rtype, body, _, loc):
            ntype = rtype
            nbody = body
            for name, atype in reversed(args):
                ntype = SAbstractionType(name, atype, ntype, loc=loc)
                nbody = SAbstraction(name, nbody, loc=loc)
            for name, kind in reversed(foralls):
                ntype = STypePolymorphism(name, kind, ntype, loc=loc)
                nbody = STypeAbstraction(name, kind, nbody, loc=loc)
            return SRec(dname, ntype, nbody, prog, loc=loc)
        case _:
            assert False, f"{d} is not a definition"


def handle_import(imp: ImportAe) -> Program:
    """Imports a given path, following the precedence rules of current folder,
    AEONPATH."""
    path = imp.path
    possible_containers = (
        [Path.cwd()] + [Path.cwd() / "libraries"] + [Path(s) for s in os.environ.get("AEONPATH", ";").split(";") if s]
    )
    for container in possible_containers:
        file = container / f"{path}"
        if file.exists():
            contents = open(file).read()
            parse = mk_parser("program", filename=str(file))
            return parse(contents)
    raise ImportError(importel=imp, possible_containers=possible_containers)



from abc import ABC
from typing import Callable

from aeon.typechecking.context import TypingContext

from aeon.core.types import Type
from aeon.core.terms import Term
from aeon.utils.name import Name

from aeon.synthesis.uis.api import SynthesisUI
from aeon.decorators.api import Metadata


# TODO add timer to synthesis
class SynthesisError(Exception):
    pass


class SynthesisNotSuccessful(SynthesisError):
    pass


class TimeoutInEvaluationException(SynthesisError):
    pass


class ErrorInSynthesis(SynthesisError):
    def __init__(self, inner_exception: Exception, msg: str):
        self.inner_exception = inner_exception
        self.msg = msg


class Synthesizer(ABC):
    def synthesize(
        self,
        ctx: TypingContext,
        type: Type,
        validate: Callable[[Term], bool],
        evaluate: Callable[[Term], list[float]],
        fun_name: Name,
        metadata: Metadata,
        budget: float = 60,
        ui: SynthesisUI = SynthesisUI(),
    ) -> Term: ...



from __future__ import annotations

from aeon.core.liquid import LiquidLiteralBool
from aeon.core.terms import (
    Abstraction,
    Annotation,
    Application,
    Hole,
    If,
    Let,
    Literal,
    Rec,
    Term,
    TypeAbstraction,
    TypeApplication,
    Var,
)
from aeon.core.types import AbstractionType
from aeon.core.types import TypeConstructor
from aeon.core.types import RefinedType
from aeon.core.types import Type
from aeon.core.types import TypeVar
from aeon.core.types import type_free_term_vars
from aeon.utils.name import Name

ops_to_abstraction: dict[str, str] = {
    "%": "Int) -> Int",
    "/": "Int) -> Int",
    "*": "Int) -> Int",
    "-": "Int) -> Int",
    "+": "Int) -> Int",
    ">=": "Int) -> Bool",
    ">": "Int) -> Bool",
    "<=": "Int) -> Bool",
    "<": "Int) -> Bool",
    "!=": "Int) -> Bool",
    "==": "Int) -> Bool",
}

aeon_prelude_ops_to_text = {
    "%": "mod",
    "/": "div",
    "*": "mul",
    "-": "sub",
    "+": "add",
    "%.": "modf",
    "/.": "divf",
    "*.": "mulf",
    "-.": "subf",
    "+.": "addf",
    ">=": "gte",
    ">": "gt",
    "<=": "lte",
    "<": "lt",
    "!=": "ne",
    "==": "eq",
    "&&": "and",
    "||": "or",
    "!": "not",
}


def pretty_print(t: Type) -> str:
    if isinstance(t, TypeConstructor):
        return str(t)
    elif isinstance(t, TypeVar):
        return str(t)
    elif isinstance(t, AbstractionType):
        at = pretty_print(t.var_type)
        rt = pretty_print(t.type)
        if t.var_name in type_free_term_vars(t.var_type):
            return f"""(
                {t.var_name}:{at}) -> {rt}"""
        else:
            return f"""{at} -> {rt}"""
    elif isinstance(t, RefinedType):
        it = pretty_print(t.type)
        if t.refinement == LiquidLiteralBool(True):
            return it
        else:
            return f"{{{t.name}:{it} | {t.refinement}}}"
    raise ValueError(f"Unknown type {t}")


def pretty_print_term(term: Term):
    term_str: str = custom_preludes_ops_representation(term)[0]
    print(term_str)


def custom_preludes_ops_representation(term: Term, counter: int = 0) -> tuple[str, int]:
    prelude_operations: dict[str, str] = aeon_prelude_ops_to_text
    match term:
        case Application(fun=Var(name=Name(var_name, _)), arg=arg) if var_name in prelude_operations.keys():
            op = var_name
            arg_str, counter = custom_preludes_ops_representation(arg, counter)
            counter += 1
            new_var_name = f"__{prelude_operations[op]}_{counter}__"
            abstraction_type_str = f"({new_var_name}:{ops_to_abstraction[op]}"
            personalized_op = f": {abstraction_type_str} = (\\{new_var_name} -> {arg_str} {op} {new_var_name})"
            return personalized_op, counter

        case Application(fun=fun, arg=arg):
            fun_str, counter = custom_preludes_ops_representation(fun, counter)
            arg_str, counter = custom_preludes_ops_representation(arg, counter)
            return (
                f"""= ({fun_str} {arg_str}
            )""",
                counter,
            )

        case Annotation(expr=expr, type=type):
            expr_str, counter = custom_preludes_ops_representation(expr, counter)
            return f"""({expr_str} : {type})""", counter

        case Abstraction(var_name=var_name, body=body):
            body_str, counter = custom_preludes_ops_representation(body, counter)
            return f"""(\\{var_name} -> {body_str})""", counter

        case Let(var_name=var_name, var_value=var_value, body=body):
            var_value_prefix = "= " if not isinstance(var_value, Application) else ""
            var_value_str, counter = custom_preludes_ops_representation(var_value, counter)
            body_str, counter = custom_preludes_ops_representation(body, counter)
            return f"""(let {var_name} {var_value_prefix}{var_value_str} in\n {body_str})""", counter

        case Rec(var_name=var_name, var_type=var_type, var_value=var_value, body=body):
            var_value_str, counter = custom_preludes_ops_representation(var_value, counter)
            body_str, counter = custom_preludes_ops_representation(body, counter)
            return f"""(let {var_name} : {var_type} = {var_value_str} in\n {body_str})""", counter

        case If(cond=cond, then=then, otherwise=otherwise):
            cond_str, counter = custom_preludes_ops_representation(cond, counter)
            then_str, counter = custom_preludes_ops_representation(then, counter)
            otherwise_str, counter = custom_preludes_ops_representation(otherwise, counter)
            return f"""(if {cond_str} then {then_str} else {otherwise_str})""", counter

        case TypeAbstraction(name=name, kind=kind, body=body):
            body_str, counter = custom_preludes_ops_representation(body, counter)
            return f"""ƛ{name}:{kind}.({body_str})""", counter

        case TypeApplication(body=body, type=type):
            body_str, counter = custom_preludes_ops_representation(body, counter)
            return f"""({body_str})[{type}]""", counter

        case Literal(_, _) | Var(_) | Hole(_):
            return str(term), counter

    return str(term), counter



from abc import ABC
from dataclasses import dataclass, field
from functools import reduce

from aeon.core.types import Kind

from typing import TYPE_CHECKING


from aeon.utils.location import Location, SynthesizedLocation
from aeon.utils.name import Name

if TYPE_CHECKING:
    from aeon.sugar.program import STerm


class SType(ABC):
    "Surface-level Type Representation"

    loc: Location


@dataclass(unsafe_hash=True)
class STypeVar(SType):
    name: Name
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"'{self.name}"


@dataclass(unsafe_hash=True)
class SRefinedType(SType):
    name: Name
    type: SType
    refinement: "STerm"
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"{{{self.name} : {self.type} | {self.refinement} }}"


@dataclass(unsafe_hash=True)
class SAbstractionType(SType):
    var_name: Name
    var_type: SType
    type: SType
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"({self.var_name} : {self.var_type}) -> {self.type}"


@dataclass(unsafe_hash=True)
class STypePolymorphism(SType):
    name: Name
    kind: Kind
    body: SType
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        return f"∀{self.name}:{self.kind}. {self.body}"


@dataclass
class STypeConstructor(SType):
    name: Name
    args: list[SType] = field(default_factory=list)
    loc: Location = field(default_factory=lambda: SynthesizedLocation("default"))

    def __str__(self):
        args = ", ".join(str(a) for a in self.args)
        return f"{self.name} {args}"

    def __hash__(self):
        return hash(self.name) + sum(hash(c) for c in self.args)


builtin_types = ["Top", "Bool", "Int", "Float", "String", "Unit"]


def get_type_vars(ty: SType) -> set[STypeVar]:
    match ty:
        case STypeVar(name):
            return {ty}
        case SAbstractionType(_, vtype, rtype):
            return get_type_vars(vtype).union(get_type_vars(rtype))
        case SRefinedType(_, rty, _):
            return get_type_vars(rty)
        case STypePolymorphism(name, _, body):
            return {t1 for t1 in get_type_vars(body) if t1.name != name}
        case STypeConstructor(name, args):
            return reduce(lambda acc, v: acc.union(get_type_vars(v)), args, set())
        case _:
            assert False, f"Unknown type ({ty}) ({type(ty)})"
